#!/usr/bin/env python3.8
# @generated by pegen from src/pegen/metagrammar.gram

import ast
import sys
import tokenize

from typing import Any, Optional

from pegen.parser import memoize, memoize_left_rec, logger, Parser
from ast import literal_eval

from pegen.grammar import (
    Alt,
    Cut,
    Forced,
    Gather,
    Group,
    Item,
    Lookahead,
    LookaheadOrCut,
    MetaTuple,
    MetaList,
    NameLeaf,
    NamedItem,
    NamedItemList,
    NegativeLookahead,
    Opt,
    Plain,
    PositiveLookahead,
    Repeat0,
    Repeat1,
    Rhs,
    Rule,
    RuleList,
    RuleName,
    Grammar,
    StringLeaf,
)


# Keywords and soft keywords are listed at the end of the parser definition.
class GeneratedParser(Parser):
    @memoize
    def start(self) -> Optional[Grammar]:
        # start: grammar $
        mark = self._mark()
        if (grammar := self.grammar()) and (self.expect("ENDMARKER")):
            return grammar
        self._reset(mark)
        return None

    @memoize
    def grammar(self) -> Optional[Grammar]:
        # grammar: metas rules | rules
        mark = self._mark()
        if (metas := self.metas()) and (rules := self.rules()):
            return Grammar(rules, metas)
        self._reset(mark)
        if rules := self.rules():
            return Grammar(rules, [])
        self._reset(mark)
        return None

    @memoize
    def metas(self) -> Optional[MetaList]:
        # metas: meta metas | meta
        mark = self._mark()
        if (meta := self.meta()) and (metas := self.metas()):
            return [meta] + metas
        self._reset(mark)
        if meta := self.meta():
            return [meta]
        self._reset(mark)
        return None

    @memoize
    def meta(self) -> Optional[MetaTuple]:
        # meta: "@" NAME NEWLINE | "@" NAME NAME NEWLINE | "@" NAME STRING NEWLINE
        mark = self._mark()
        if (self.expect("@")) and (name := self.name()) and (self.expect("NEWLINE")):
            return (name.string, None)
        self._reset(mark)
        if (
            (self.expect("@"))
            and (a := self.name())
            and (b := self.name())
            and (self.expect("NEWLINE"))
        ):
            return (a.string, b.string)
        self._reset(mark)
        if (
            (self.expect("@"))
            and (name := self.name())
            and (string := self.string())
            and (self.expect("NEWLINE"))
        ):
            return (name.string, literal_eval(string.string))
        self._reset(mark)
        return None

    @memoize
    def rules(self) -> Optional[RuleList]:
        # rules: rule rules | rule
        mark = self._mark()
        if (rule := self.rule()) and (rules := self.rules()):
            return [rule] + rules
        self._reset(mark)
        if rule := self.rule():
            return [rule]
        self._reset(mark)
        return None

    @memoize
    def rule(self) -> Optional[Rule]:
        # rule: rulename memoflag? ":" alts NEWLINE INDENT more_alts DEDENT | rulename memoflag? ":" NEWLINE INDENT more_alts DEDENT | rulename memoflag? ":" alts NEWLINE
        mark = self._mark()
        if (
            (rulename := self.rulename())
            and (opt := self.memoflag(),)
            and (self.expect(":"))
            and (alts := self.alts())
            and (self.expect("NEWLINE"))
            and (self.expect("INDENT"))
            and (more_alts := self.more_alts())
            and (self.expect("DEDENT"))
        ):
            return Rule(rulename[0], rulename[1], Rhs(alts.alts + more_alts.alts), memo=opt)
        self._reset(mark)
        if (
            (rulename := self.rulename())
            and (opt := self.memoflag(),)
            and (self.expect(":"))
            and (self.expect("NEWLINE"))
            and (self.expect("INDENT"))
            and (more_alts := self.more_alts())
            and (self.expect("DEDENT"))
        ):
            return Rule(rulename[0], rulename[1], more_alts, memo=opt)
        self._reset(mark)
        if (
            (rulename := self.rulename())
            and (opt := self.memoflag(),)
            and (self.expect(":"))
            and (alts := self.alts())
            and (self.expect("NEWLINE"))
        ):
            return Rule(rulename[0], rulename[1], alts, memo=opt)
        self._reset(mark)
        return None

    @memoize
    def rulename(self) -> Optional[RuleName]:
        # rulename: NAME annotation | NAME
        mark = self._mark()
        if (name := self.name()) and (annotation := self.annotation()):
            return (name.string, annotation)
        self._reset(mark)
        if name := self.name():
            return (name.string, None)
        self._reset(mark)
        return None

    @memoize
    def memoflag(self) -> Optional[str]:
        # memoflag: '(' "memo" ')'
        mark = self._mark()
        if (self.expect("(")) and (self.expect("memo")) and (self.expect(")")):
            return "memo"
        self._reset(mark)
        return None

    @memoize
    def alts(self) -> Optional[Rhs]:
        # alts: alt "|" alts | alt
        mark = self._mark()
        if (alt := self.alt()) and (self.expect("|")) and (alts := self.alts()):
            return Rhs([alt] + alts.alts)
        self._reset(mark)
        if alt := self.alt():
            return Rhs([alt])
        self._reset(mark)
        return None

    @memoize
    def more_alts(self) -> Optional[Rhs]:
        # more_alts: "|" alts NEWLINE more_alts | "|" alts NEWLINE
        mark = self._mark()
        if (
            (self.expect("|"))
            and (alts := self.alts())
            and (self.expect("NEWLINE"))
            and (more_alts := self.more_alts())
        ):
            return Rhs(alts.alts + more_alts.alts)
        self._reset(mark)
        if (self.expect("|")) and (alts := self.alts()) and (self.expect("NEWLINE")):
            return Rhs(alts.alts)
        self._reset(mark)
        return None

    @memoize
    def alt(self) -> Optional[Alt]:
        # alt: items '$' action | items '$' | items action | items
        mark = self._mark()
        if (items := self.items()) and (self.expect("$")) and (action := self.action()):
            return Alt(items + [NamedItem(None, NameLeaf("ENDMARKER"))], action=action)
        self._reset(mark)
        if (items := self.items()) and (self.expect("$")):
            return Alt(items + [NamedItem(None, NameLeaf("ENDMARKER"))], action=None)
        self._reset(mark)
        if (items := self.items()) and (action := self.action()):
            return Alt(items, action=action)
        self._reset(mark)
        if items := self.items():
            return Alt(items, action=None)
        self._reset(mark)
        return None

    @memoize
    def items(self) -> Optional[NamedItemList]:
        # items: named_item items | named_item
        mark = self._mark()
        if (named_item := self.named_item()) and (items := self.items()):
            return [named_item] + items
        self._reset(mark)
        if named_item := self.named_item():
            return [named_item]
        self._reset(mark)
        return None

    @memoize
    def named_item(self) -> Optional[NamedItem]:
        # named_item: NAME annotation '=' ~ item | NAME '=' ~ item | item | forced_atom | lookahead
        mark = self._mark()
        cut = False
        if (
            (name := self.name())
            and (annotation := self.annotation())
            and (self.expect("="))
            and (cut := True)
            and (item := self.item())
        ):
            return NamedItem(name.string, item, annotation)
        self._reset(mark)
        if cut:
            return None
        cut = False
        if (
            (name := self.name())
            and (self.expect("="))
            and (cut := True)
            and (item := self.item())
        ):
            return NamedItem(name.string, item)
        self._reset(mark)
        if cut:
            return None
        if item := self.item():
            return NamedItem(None, item)
        self._reset(mark)
        if it := self.forced_atom():
            return NamedItem(None, it)
        self._reset(mark)
        if it := self.lookahead():
            return NamedItem(None, it)
        self._reset(mark)
        return None

    @memoize
    def forced_atom(self) -> Optional[LookaheadOrCut]:
        # forced_atom: '&' '&' ~ atom
        mark = self._mark()
        cut = False
        if (self.expect("&")) and (self.expect("&")) and (cut := True) and (atom := self.atom()):
            return Forced(atom)
        self._reset(mark)
        if cut:
            return None
        return None

    @memoize
    def lookahead(self) -> Optional[LookaheadOrCut]:
        # lookahead: '&' ~ atom | '!' ~ atom | '~'
        mark = self._mark()
        cut = False
        if (self.expect("&")) and (cut := True) and (atom := self.atom()):
            return PositiveLookahead(atom)
        self._reset(mark)
        if cut:
            return None
        cut = False
        if (self.expect("!")) and (cut := True) and (atom := self.atom()):
            return NegativeLookahead(atom)
        self._reset(mark)
        if cut:
            return None
        if self.expect("~"):
            return Cut()
        self._reset(mark)
        return None

    @memoize
    def item(self) -> Optional[Item]:
        # item: '[' ~ alts ']' | atom '?' | atom '*' | atom '+' | atom '.' atom '+' | atom
        mark = self._mark()
        cut = False
        if (self.expect("[")) and (cut := True) and (alts := self.alts()) and (self.expect("]")):
            return Opt(alts)
        self._reset(mark)
        if cut:
            return None
        if (atom := self.atom()) and (self.expect("?")):
            return Opt(atom)
        self._reset(mark)
        if (atom := self.atom()) and (self.expect("*")):
            return Repeat0(atom)
        self._reset(mark)
        if (atom := self.atom()) and (self.expect("+")):
            return Repeat1(atom)
        self._reset(mark)
        if (
            (sep := self.atom())
            and (self.expect("."))
            and (node := self.atom())
            and (self.expect("+"))
        ):
            return Gather(sep, node)
        self._reset(mark)
        if atom := self.atom():
            return atom
        self._reset(mark)
        return None

    @memoize
    def atom(self) -> Optional[Plain]:
        # atom: '(' ~ alts ')' | NAME | STRING
        mark = self._mark()
        cut = False
        if (self.expect("(")) and (cut := True) and (alts := self.alts()) and (self.expect(")")):
            return Group(alts)
        self._reset(mark)
        if cut:
            return None
        if name := self.name():
            return NameLeaf(name.string)
        self._reset(mark)
        if string := self.string():
            return StringLeaf(string.string)
        self._reset(mark)
        return None

    @memoize
    def action(self) -> Optional[str]:
        # action: "{" ~ target_atoms "}"
        mark = self._mark()
        cut = False
        if (
            (self.expect("{"))
            and (cut := True)
            and (target_atoms := self.target_atoms())
            and (self.expect("}"))
        ):
            return target_atoms
        self._reset(mark)
        if cut:
            return None
        return None

    @memoize
    def annotation(self) -> Optional[str]:
        # annotation: "[" ~ target_atoms "]"
        mark = self._mark()
        cut = False
        if (
            (self.expect("["))
            and (cut := True)
            and (target_atoms := self.target_atoms())
            and (self.expect("]"))
        ):
            return target_atoms
        self._reset(mark)
        if cut:
            return None
        return None

    @memoize
    def target_atoms(self) -> Optional[str]:
        # target_atoms: target_atom target_atoms | target_atom
        mark = self._mark()
        if (target_atom := self.target_atom()) and (target_atoms := self.target_atoms()):
            return target_atom + " " + target_atoms
        self._reset(mark)
        if target_atom := self.target_atom():
            return target_atom
        self._reset(mark)
        return None

    @memoize
    def target_atom(self) -> Optional[str]:
        # target_atom: "{" ~ target_atoms? "}" | "[" ~ target_atoms? "]" | NAME "*" | NAME | NUMBER | STRING | FSTRING_START target_fstring_middle* FSTRING_END | "?" | ":" | !"}" !"]" OP
        mark = self._mark()
        cut = False
        if (
            (self.expect("{"))
            and (cut := True)
            and (atoms := self.target_atoms(),)
            and (self.expect("}"))
        ):
            return "{" + (atoms or "") + "}"
        self._reset(mark)
        if cut:
            return None
        cut = False
        if (
            (self.expect("["))
            and (cut := True)
            and (atoms := self.target_atoms(),)
            and (self.expect("]"))
        ):
            return "[" + (atoms or "") + "]"
        self._reset(mark)
        if cut:
            return None
        if (name := self.name()) and (self.expect("*")):
            return name.string + "*"
        self._reset(mark)
        if name := self.name():
            return name.string
        self._reset(mark)
        if number := self.number():
            return number.string
        self._reset(mark)
        if string := self.string():
            return string.string
        self._reset(mark)
        if (l := self.fstring_start()) and (m := self._loop0_1(),) and (r := self.fstring_end()):
            return l.string + "".join(m) + r.string
        self._reset(mark)
        if self.expect("?"):
            return "?"
        self._reset(mark)
        if self.expect(":"):
            return ":"
        self._reset(mark)
        if (
            (self.negative_lookahead(self.expect, "}"))
            and (self.negative_lookahead(self.expect, "]"))
            and (op := self.op())
        ):
            return op.string
        self._reset(mark)
        return None

    @memoize
    def target_fstring_middle(self) -> Optional[str]:
        # target_fstring_middle: FSTRING_MIDDLE | "{" | "}" | target_atom
        mark = self._mark()
        if fstring_middle := self.fstring_middle():
            return fstring_middle.string
        self._reset(mark)
        if self.expect("{"):
            return "{"
        self._reset(mark)
        if self.expect("}"):
            return "}"
        self._reset(mark)
        if target_atom := self.target_atom():
            return target_atom
        self._reset(mark)
        return None

    @memoize
    def _loop0_1(self) -> Optional[Any]:
        # _loop0_1: target_fstring_middle
        mark = self._mark()
        children = []
        while target_fstring_middle := self.target_fstring_middle():
            children.append(target_fstring_middle)
            mark = self._mark()
        self._reset(mark)
        return children

    KEYWORDS = ()
    SOFT_KEYWORDS = ("memo",)


if __name__ == "__main__":
    from pegen.parser import simple_parser_main

    simple_parser_main(GeneratedParser)



"""Test syntax errors for cases where the parser can generate helpful messages."""
import sys

import pytest


def parse_invalid_syntax(
    python_parse_file,
    python_parse_str,
    tmp_path,
    source,
    exc_cls,
    message,
    start,
    end,
    min_python_version=(3, 10),
) -> None:
    # Check we obtain the expected error from Python
    try:
        exec(source, {}, {})
    except exc_cls as py_e:
        py_exc = py_e
    except Exception as py_e:
        assert (
            False
        ), f"Python produced {py_e.__class__.__name__} instead of {exc_cls.__name__}: {py_e}"
    else:
        assert False, f"Python did not throw any exception, expected {exc_cls}"

    # Check our parser raises both from str and file mode.
    with pytest.raises(exc_cls) as e:
        python_parse_str(source, "exec")

    print(str(e.exconly()))
    assert message in str(e.exconly())

    test_file = tmp_path / "test.py"
    with open(test_file, "w") as f:
        f.write(source)

    with pytest.raises(exc_cls) as e:
        python_parse_file(str(test_file))

    # Check Python message but do not expect message to match for earlier Python versions
    if sys.version_info >= min_python_version:
        # This fails for Python < 3.10.5 but keeping the fix for a patch version is not
        # worth it
        assert message in py_exc.args[0]

    print(str(e.exconly()))
    assert message in str(e.exconly())

    if start is None:
        return

    # Check start/end line/column on Python 3.10
    for parser, exc in ([("Python", py_exc)] if sys.version_info >= min_python_version else []) + [
        ("pegen", e.value)
    ]:
        if (
            exc.lineno != start[0]
            or exc.offset != start[1]
            # Do not check end for indentation errors
            or (
                sys.version_info >= (3, 10)
                and not isinstance(e, IndentationError)
                and exc.end_lineno != end[0]
            )
            or (
                sys.version_info >= (3, 10)
                and not isinstance(e, IndentationError)
                and (end[1] is not None and exc.end_offset != end[1])
            )
        ):
            if sys.version_info >= (3, 10):
                raise ValueError(
                    f"Expected locations of {start} and {end}, but got "
                    f"{(exc.lineno, exc.offset)} and {(exc.end_lineno, exc.end_offset)} "
                    f"from {parser}"
                )
            else:
                raise ValueError(
                    f"Expected locations of {start}, but got "
                    f"{(exc.lineno, exc.offset)} from {parser}"
                )


@pytest.mark.skipif(
    sys.version_info < (3, 9), reason="Python 3.8 diagnostic is subpar in this case."
)
@pytest.mark.parametrize(
    "source, message, start, end",
    [
        (
            "f'a = {}'",
            "valid expression required before '}'"
            if sys.version_info >= (3, 12)
            else "f-string: empty expression not allowed",
            (1, 8) if sys.version_info >= (3, 12) else None,
            (1, 9) if sys.version_info >= (3, 12) else None,
        ),
        (
            "f'a = {=}'",
            "expression required before '='"
            if sys.version_info >= (3, 11)
            else "f-string: empty expression not allowed",
            (1, 8) if sys.version_info >= (3, 12) else None,
            (1, 9) if sys.version_info >= (3, 12) else None,
        ),
        (
            "f'a = {!}'",
            "expression required before '!'"
            if sys.version_info >= (3, 11)
            else "f-string: empty expression not allowed",
            (1, 8) if sys.version_info >= (3, 12) else None,
            (1, 9) if sys.version_info >= (3, 12) else None,
        ),
        (
            "f'a = {:}'",
            "expression required before ':'"
            if sys.version_info >= (3, 11)
            else "f-string: empty expression not allowed",
            (1, 8) if sys.version_info >= (3, 12) else None,
            (1, 9) if sys.version_info >= (3, 12) else (1, 11),
        ),
        (
            "f'a = {a=d}'",
            "expecting '!', or ':', or '}'"
            if sys.version_info >= (3, 12)
            else "f-string: expecting '}'",
            (1, 10) if sys.version_info >= (3, 12) else None,
            (1, 11) if sys.version_info >= (3, 12) else None,
        ),
        (
            "f'a = { 1 + }'",
            "f-string: expecting '=', or '!', or ':', or '}'"
            if sys.version_info >= (3, 12)
            else "f-string: invalid syntax",
            (1, 11) if sys.version_info >= (3, 12) else (1, 7),
            (1, 12) if sys.version_info >= (3, 12) else (1, 8),
        ),
        (
            "(\n\t'b'\n\tf'a = { 1 + }'\n)",
            "f-string: expecting '=', or '!', or ':', or '}'"
            if sys.version_info >= (3, 12)
            else "f-string: invalid syntax",
            (3, 12) if sys.version_info >= (3, 12) else (3, 7),
            (3, 13) if sys.version_info >= (3, 12) else (3, 8),
        ),
    ],
)
def test_syntax_error_in_str(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("(a 1)", "invalid syntax. Perhaps you forgot a comma?", (1, 2), (1, 5)),
        ("2 if 4", "expected 'else' after 'if' expression", (1, 1), (1, 7)),
        ("(a 1 if b else 2)", "invalid syntax. Perhaps you forgot a comma?", (1, 2), (1, 17)),
        ("(a lambda: 1)", "invalid syntax. Perhaps you forgot a comma?", (1, 2), (1, 13)),
        ("(a b)", "invalid syntax. Perhaps you forgot a comma?", (1, 2), (1, 5)),
        ("print 1", "Missing parentheses in call to 'print'", (1, 1), (1, 8)),
        ("exec 1", "Missing parentheses in call to 'exec'", (1, 1), (1, 7)),
        ("a if b", "expected 'else' after 'if' expression", (1, 1), (1, 7)),
        ("c = a if b:", "invalid syntax", (1, 11), (1, 12)),
    ],
)
def test_invalid_expression(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


# Those tests are mostly there to get coverage on exiting rules without matching
@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("global a, 1", "invalid syntax", (1, 11), (1, 12)),
        ("nonlocal a, 1", "invalid syntax", (1, 13), (1, 14)),
        ("yield raise", "invalid syntax", (1, 7), (1, 12)),
        ("assert raise", "invalid syntax", (1, 8), (1, 13)),
        ("return def", "invalid syntax", (1, 8), (1, 11)),
        ("raise def", "invalid syntax", (1, 7), (1, 10)),
        ("del raise", "invalid syntax", (1, 5), (1, 10)),
        ("if raise:\n\tpass", "invalid syntax", (1, 4), (1, 9)),
        ("@raise\ndef f():\n\tpass", "invalid syntax", (1, 2), (1, 7)),
        ("a: int = raise", "invalid syntax", (1, 10), (1, 15)),
    ],
)
def test_invalid_statements(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        # Invalid arguments rules
        (
            "f(**a, *b)",
            "iterable argument unpacking follows keyword argument unpacking",
            (1, 3),
            (1, 6),
        ),
        # NOTE CPython bug, should report 15 as expected (we use None to omit the check)
        ("f(a for a in b, c)", "Generator expression must be parenthesized", (1, 3), (1, None)),
        # NOTE CPython bug, should report 20 as expected (we use None to omit the check)
        (
            "f(a for a in b if a, c)",
            "Generator expression must be parenthesized",
            (1, 3),
            (1, None),
        ),
        # NOTE CPython bug, should report 15 as expected (we use None to omit the check)
        (
            "f(a for a in b, c for c in d)",
            "Generator expression must be parenthesized",
            (1, 3),
            (1, None),
        ),
        (
            "f(a=1 for i in range(10))",
            "invalid syntax. Maybe you meant '==' or ':=' instead of '='?",
            (1, 3),
            (1, 5),
        ),
        # NOTE CPython bug, should report 18 as expected (we use None to omit the check)
        ("f(a, b for b in c)", "Generator expression must be parenthesized", (1, 6), (1, None)),
        # NOTE CPython bug, should report 18 as expected (we use None to omit the check)
        ("f(a, b for b in c, d)", "Generator expression must be parenthesized", (1, 6), (1, None)),
        ("f(*a=b)", "cannot assign to iterable argument unpacking", (1, 3), (1, 7)),
        ("f(**a, b)", "positional argument follows keyword argument unpacking", (1, 9), (1, 10)),
        ("f(a=1, b)", "positional argument follows keyword argument", (1, 9), (1, 10)),
        # Invalid kwarg rules
        (
            "f(b=c for c in d)",
            "invalid syntax. Maybe you meant '==' or ':=' instead of '='?",
            (1, 3),
            (1, 5),
        ),
        (
            "f(1 + b=2)",
            'expression cannot contain assignment, perhaps you meant "=="?',
            (1, 3),
            (1, 9),
        ),
        ("f(True=1)", "cannot assign to True", (1, 3), (1, 8)),
        ("f(False=1)", "cannot assign to False", (1, 3), (1, 9)),
        ("f(None=1)", "cannot assign to None", (1, 3), (1, 8)),
    ],
)
def test_invalid_call_arguments(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file,
        python_parse_str,
        tmp_path,
        source,
        SyntaxError,
        message,
        start,
        end,
        (3, 12) if "cannot assign" in message else (3, 10),
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("'a' = 1", "cannot assign to literal", (1, 1), (1, 4)),
        ("1 = 1", "cannot assign to literal", (1, 1), (1, 2)),
        ("True = 1", "cannot assign to True", (1, 1), (1, 5)),
        ("False = 1", "cannot assign to False", (1, 1), (1, 6)),
        ("... = 1", "cannot assign to ellipsis", (1, 1), (1, 4)),
        ("None = 1", "cannot assign to None", (1, 1), (1, 5)),
        (
            "(a, b) : int = (1, 2)",
            "only single target (not tuple) can be annotated",
            (1, 1),
            (1, 7),
        ),
        (
            "[a, b] : int = [1, 2]",
            "only single target (not list) can be annotated",
            (1, 1),
            (1, 7),
        ),
        (
            "([a, b]) : int = (1, 2)",
            "only single target (not list) can be annotated",
            (1, 2),
            (1, 8),
        ),
        (
            "a, b: int, int = 1, 2",
            "only single target (not tuple) can be annotated",
            (1, 1),
            (1, 2),
        ),
        ("{a, b} : set", "illegal target for annotation", (1, 1), (1, 7)),
        ("a + 1 = 2", "cannot assign to expression", (1, 1), (1, 6)),
        ("[i for i in range(2)] = 2", "cannot assign to list comprehension", (1, 1), (1, 22)),
        ("yield a = 1", "assignment to yield expression not possible", (1, 1), (1, 8)),
        ("a = yield b = 1", "assignment to yield expression not possible", (1, 5), (1, 12)),
        (
            "a + 1 += 1",
            "'expression' is an illegal expression for augmented assignment",
            (1, 1),
            (1, 6),
        ),
        (
            "a + 1 += yield",
            "'expression' is an illegal expression for augmented assignment",
            (1, 1),
            (1, 6),
        ),
        (
            "[i for i in range(2)] += 1",
            "'list comprehension' is an illegal expression for augmented assignment",
            (1, 1),
            (1, 22),
        ),
        ("a += raise", "invalid syntax", (1, 6), (1, 11)),
    ],
)
def test_invalid_assignments(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("del [i for i in range(2)]", "cannot delete list comprehension", (1, 5), (1, 26)),
        ("del a + 1", "cannot delete expression", (1, 5), (1, 10)),
    ],
)
def test_invalid_del_statements(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        (
            "(*a for a in enumerate(range(5)))",
            "iterable unpacking cannot be used in comprehension",
            (1, 2),
            (1, 4),
        ),
        (
            "[*a for a in enumerate(range(5))]",
            "iterable unpacking cannot be used in comprehension",
            (1, 2),
            (1, 4),
        ),
        (
            "{*a for a in enumerate(range(5))}",
            "iterable unpacking cannot be used in comprehension",
            (1, 2),
            (1, 4),
        ),
        (
            "[a, a for a in range(5)]",
            "did you forget parentheses around the comprehension target?",
            (1, 2),
            (1, 6),
        ),
        (
            "[a, a for a in range(5)]",
            "did you forget parentheses around the comprehension target?",
            (1, 2),
            (1, 6),
        ),
        (
            "[a,  for a in range(5)]",
            "did you forget parentheses around the comprehension target?",
            (1, 2),
            (1, 4),
        ),
        (
            "[a,  for a in range(5)]",
            "did you forget parentheses around the comprehension target?",
            (1, 2),
            (1, 4),
        ),
        (
            "{**a for a in [{1: 2}]}",
            "dict unpacking cannot be used in dict comprehension",
            (1, 2),
            (1, 4),
        ),
        # check cuts
        ("(a for a in raise)", "invalid syntax", (1, 13), (1, 18)),
        ("(a async for a in raise)", "invalid syntax", (1, 19), (1, 24)),
    ],
)
def test_invalid_comprehension(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end, py_version",
    [
        (
            "def f(a=1, b):\n\tpass",
            "parameter without a default follows parameter with a default",
            (1, 12),
            (1, 13),
            (3, 12),  # Error message was improved in 3.12
        ),
        (
            "def f(a=1, /, b):\n\tpass",
            "parameter without a default follows parameter with a default",
            (1, 15),
            (1, 16),
            (3, 12),  # Error message was improved in 3.12
        ),
        (
            "def f(x, (y, z), w):\n\tpass",
            "parameters cannot be parenthesized",
            (1, 10),
            (1, 16),
            (3, 11),
        ),
        (
            "def f((x, y, z, w)):\n\tpass",
            "parameters cannot be parenthesized",
            (1, 7),
            (1, 19),
            (3, 11),
        ),
        (
            "def f(x, (y, z, w)):\n\tpass",
            "parameters cannot be parenthesized",
            (1, 10),
            (1, 19),
            (3, 11),
        ),
        (
            "def f((x, y, z), w):\n\tpass",
            "parameters cannot be parenthesized",
            (1, 7),
            (1, 16),
            (3, 11),
        ),
        (
            "def foo(/,a,b=,c):\n\tpass",
            "at least one argument must precede /",
            (1, 9),
            (1, 10),
            (3, 11),
        ),
        ("def foo(a,/,/,b,c):\n\tpass", "/ may appear only once", (1, 13), (1, 14), (3, 11)),
        ("def foo(a,/,a1,/,b,c):\n\tpass", "/ may appear only once", (1, 16), (1, 17), (3, 11)),
        ("def foo(a=1,/,/,*b,/,c):\n\tpass", "/ may appear only once", (1, 15), (1, 16), (3, 11)),
        ("def foo(a,/,a1=1,/,b,c):\n\tpass", "/ may appear only once", (1, 18), (1, 19), (3, 11)),
        ("def foo(a,*b,c,/,d,e):\n\tpass", "/ must be ahead of *", (1, 16), (1, 17), (3, 11)),
        ("def foo(a=1,*b,c=3,/,d,e):\n\tpass", "/ must be ahead of *", (1, 20), (1, 21), (3, 11)),
        (
            "def foo(a,*b=3,c):\n\tpass",
            "var-positional argument cannot have default value",
            (1, 13),
            (1, 14),
            (3, 11),
        ),
        (
            "def foo(a,*b: int=,c):\n\tpass",
            "var-positional argument cannot have default value",
            (1, 18),
            (1, 19),
            (3, 11),
        ),
        (
            "def foo(a,**b=3):\n\tpass",
            "var-keyword argument cannot have default value",
            (1, 14),
            (1, 15),
            (3, 11),
        ),
        (
            "def foo(a,**b: int=3):\n\tpass",
            "var-keyword argument cannot have default value",
            (1, 19),
            (1, 20),
            (3, 11),
        ),
        (
            "def foo(a,*a, b, **c, d):\n\tpass",
            "arguments cannot follow var-keyword argument",
            (1, 23),
            (1, 24),
            (3, 11),
        ),
        (
            "def foo(a,*a, b, **c, d=4):\n\tpass",
            "arguments cannot follow var-keyword argument",
            (1, 23),
            (1, 24),
            (3, 11),
        ),
        (
            "def foo(a,*a, b, **c, *d):\n\tpass",
            "arguments cannot follow var-keyword argument",
            (1, 23),
            (1, 24),
            (3, 11),
        ),
        (
            "def foo(a,*a, b, **c, **d):\n\tpass",
            "arguments cannot follow var-keyword argument",
            (1, 23),
            (1, 25),
            (3, 11),
        ),
        (
            "def foo(a=1,/,**b,/,c):\n\tpass",
            "arguments cannot follow var-keyword argument",
            (1, 19),
            (1, 20),
            (3, 11),
        ),
        ("def foo(*b,*d):\n\tpass", "* argument may appear only once", (1, 12), (1, 13), (3, 11)),
        (
            "def foo(a,*b,c,*d,*e,c):\n\tpass",
            "* argument may appear only once",
            (1, 16),
            (1, 17),
            (3, 11),
        ),
        (
            "def foo(a,b,/,c,*b,c,*d,*e,c):\n\tpass",
            "* argument may appear only once",
            (1, 22),
            (1, 23),
            (3, 11),
        ),
        (
            "def foo(a,b,/,c,*b,c,*d,**e):\n\tpass",
            "* argument may appear only once",
            (1, 22),
            (1, 23),
            (3, 11),
        ),
        (
            "def foo(a=1,/*,b,c):\n\tpass",
            "expected comma between / and *",
            (1, 14),
            (1, 15),
            (3, 11),
        ),
        (
            "def foo(a=1,d=,c):\n\tpasss",
            "expected default value expression",
            (1, 14),
            (1, 15),
            (3, 11),
        ),
        (
            "def foo(a,d=,c):\n\tpass",
            "expected default value expression",
            (1, 12),
            (1, 13),
            (3, 11),
        ),
        (
            "def foo(a,d: int=,c):\n\tpass",
            "expected default value expression",
            (1, 17),
            (1, 18),
            (3, 11),
        ),
        (
            "lambda x=1, y: x",
            "parameter without a default follows parameter with a default",
            (1, 13),
            (1, 14),
            (3, 12),  # Improved error message
        ),
        (
            "lambda x=1, /, y: x",
            "parameter without a default follows parameter with a default",
            (1, 16),
            (1, 17),
            (3, 12),  # improved error message
        ),
        (
            "lambda x, (y, z), w: None",
            "parameters cannot be parenthesized",
            (1, 11),
            (1, 17),
            (3, 11),
        ),
        (
            "lambda (x, y, z, w): None",
            "parameters cannot be parenthesized",
            (1, 8),
            (1, 20),
            (3, 11),
        ),
        (
            "lambda x, (y, z, w): None",
            "parameters cannot be parenthesized",
            (1, 11),
            (1, 20),
            (3, 11),
        ),
        (
            "lambda (x, y, z), w: None",
            "parameters cannot be parenthesized",
            (1, 8),
            (1, 17),
            (3, 11),
        ),
        ("lambda /,a,b,c: None", "at least one argument must precede /", (1, 8), (1, 9), (3, 11)),
        ("lambda a,/,/,b,c: None", "/ may appear only once", (1, 12), (1, 13), (3, 11)),
        ("lambda a,/,a1,/,b,c: None", "/ may appear only once", (1, 15), (1, 16), (3, 11)),
        ("lambda a=1,/,/,*b,/,c: None", "/ may appear only once", (1, 14), (1, 15), (3, 11)),
        ("lambda a,/,a1=1,/,b,c: None", "/ may appear only once", (1, 17), (1, 18), (3, 11)),
        ("lambda a,*b,c,/,d,e: None", "/ must be ahead of *", (1, 15), (1, 16), (3, 11)),
        ("lambda a=1,*b,c=3,/,d,e: None", "/ must be ahead of *", (1, 19), (1, 20), (3, 11)),
        ("lambda a=1,/*,b,c: None", "expected comma between / and *", (1, 13), (1, 14), (3, 11)),
        (
            "lambda a,*b=3,c: None",
            "var-positional argument cannot have default value",
            (1, 12),
            (1, 13),
            (3, 11),
        ),
        (
            "lambda a,**b=3: None",
            "var-keyword argument cannot have default value",
            (1, 13),
            (1, 14),
            (3, 11),
        ),
        (
            "lambda a, *a, b, **c, d: None",
            "arguments cannot follow var-keyword argument",
            (1, 23),
            (1, 24),
            (3, 11),
        ),
        (
            "lambda a,*a, b, **c, d=4: None",
            "arguments cannot follow var-keyword argument",
            (1, 22),
            (1, 23),
            (3, 11),
        ),
        (
            "lambda a,*a, b, **c, *d: None",
            "arguments cannot follow var-keyword argument",
            (1, 22),
            (1, 23),
            (3, 11),
        ),
        (
            "lambda a,*a, b, **c, **d: None",
            "arguments cannot follow var-keyword argument",
            (1, 22),
            (1, 24),
            (3, 11),
        ),
        (
            "lambda a=1,/,**b,/,c: None",
            "arguments cannot follow var-keyword argument",
            (1, 18),
            (1, 19),
            (3, 11),
        ),
        ("lambda *b,*d: None", "* argument may appear only once", (1, 11), (1, 12), (3, 11)),
        (
            "lambda a,*b,c,*d,*e,c: None",
            "* argument may appear only once",
            (1, 15),
            (1, 16),
            (3, 11),
        ),
        (
            "lambda a,b,/,c,*b,c,*d,*e,c: None",
            "* argument may appear only once",
            (1, 21),
            (1, 22),
            (3, 11),
        ),
        (
            "lambda a,b,/,c,*b,c,*d,**e: None",
            "* argument may appear only once",
            (1, 21),
            (1, 22),
            (3, 11),
        ),
        ("lambda a=1,d=,c: None", "expected default value expression", (1, 13), (1, 14), (3, 11)),
        ("lambda a,d=,c: None", "expected default value expression", (1, 11), (1, 12), (3, 11)),
    ],
)
def test_invalid_parameters(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end, py_version
):
    parse_invalid_syntax(
        python_parse_file,
        python_parse_str,
        tmp_path,
        source,
        SyntaxError,
        message,
        start,
        end,
        py_version,
    )


@pytest.mark.parametrize(
    "source, start, end",
    [
        ("def f(a, *):\n\tpass", (1, 10), (1, 11)),
        ("def f(a, *,):\n\tpass", (1, 10), (1, 11)),
        ("def f(a, *, **):\n\tpass", (1, 10), (1, 11)),
        ("lambda a, *: a", (1, 12), (1, 13)),
        ("lambda a, *, **:a", (1, 14), (1, 16)),
    ],
)
def test_invalid_star_etc(python_parse_file, python_parse_str, tmp_path, source, start, end):
    parse_invalid_syntax(
        python_parse_file,
        python_parse_str,
        tmp_path,
        source,
        SyntaxError,
        "named arguments must follow bare *",
        start,
        end,
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("with open(a) as {b: 1}:\n\tpass", "cannot assign to dict", (1, 17), (1, 23)),
        ("with open(a) as {b}:\n\tpass", "cannot assign to set", (1, 17), (1, 20)),
        ("with open(a) as 1:\n\tpass", "cannot assign to literal", (1, 17), (1, 18)),
    ],
)
def test_invalid_with_item(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("for {a} in [[1]]:\n\tpass", "cannot assign to set display", (1, 5), (1, 8)),
        ("async for (a := i)", "cannot assign to named expression", (1, 12), (1, 18)),
    ],
)
def test_invalid_for_target(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("a = (1+1 := 2)", "cannot use assignment expressions with expression", (1, 6), (1, 9)),
        ("a := raise", "invalid syntax", (1, 3), (1, 5)),
    ],
)
def test_named_expression(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        ("a = (*b)", "cannot use starred expression here", (1, 6), (1, 8)),
        ("a = (**b)", "cannot use double starred expression here", (1, 6), (1, 8)),
    ],
)
def test_invalid_group(python_parse_file, python_parse_str, tmp_path, source, message, start, end):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, message, start, end",
    [
        (
            "from a import b,",
            "trailing comma not allowed without surrounding parentheses",
            (1, 17),
            (1, 18) if sys.version_info >= (3, 12) else (1, 17),
        ),
        ("from a import b, and 3", "invalid syntax", (1, 18), (1, 21)),
        ("from a import raise", "invalid syntax", (1, 15), (1, 20)),
        pytest.param(
            "import a from b",
            "'from ... import ...'",
            (1, 1),
            (1, 16),
            marks=pytest.mark.skipif(sys.version_info < (3, 12), reason="Requires Python 3.12+"),
        ),
    ],
)
def test_invalid_import_from_as_names(
    python_parse_file, python_parse_str, tmp_path, source, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, SyntaxError, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "with open(a) as f, b as d:\npass",
            IndentationError,
            "expected an indented block after 'with' statement on line 1",
            (2, 1),
            (2, 5),
        ),
        # Python 3.9+ only
        pytest.param(
            "\nasync with (open(a) as f, b as d):\npass",
            IndentationError,
            "expected an indented block after 'with' statement on line 2",
            (3, 1),
            (3, 5),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 9), reason="Unsupported syntax on Python 3.8"
            ),
        ),
        (
            "with open(a) as f, b as d\npass",
            SyntaxError,
            "expected ':'",
            (1, 26),
            (1, 27) if sys.version_info >= (3, 12) else (1, 26),
        ),
        (
            "\nasync with (open(a) as f, b as d)\npass",
            SyntaxError,
            "expected ':'",
            (2, 34),
            (2, 35) if sys.version_info >= (3, 12) else (2, 34),
        ),
    ],
)
def test_invalid_with_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "try:\npass",
            IndentationError,
            "expected an indented block after 'try' statement on line 1",
            (2, 1),
            (2, 5),
        ),
        (
            "try\n\tpass",
            SyntaxError,
            "expected ':'",
            (1, 4),
            (1, 5) if sys.version_info >= (3, 12) else (1, 4),
        ),
        (
            "try:\n\tpass\na = 1",
            SyntaxError,
            "expected 'except' or 'finally' block",
            (3, 1),
            (3, 2),
        ),
    ],
)
def test_invalid_try_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "try:\n\tpass\nexcept:\npass",
            # NOTE: there are 2 CPython bugs here, the exception type and the missing quote
            # in the message
            SyntaxError,  # IndentationError,
            "expected an indented block after 'except' statement on line 3",
            (4, 1),
            (4, 5),
        ),
        (
            "try:\n\tpass\nexcept Exception:\npass",
            IndentationError,
            "expected an indented block after 'except' statement on line 3",
            (4, 1),
            (4, 5),
        ),
        (
            "try:\n\tpass\nexcept Exception as e:\npass",
            IndentationError,
            "expected an indented block after 'except' statement on line 3",
            (4, 1),
            (4, 5),
        ),
        pytest.param(
            "try:\n\tpass\nexcept* Exception as e:\npass",
            IndentationError,
            "expected an indented block after 'except*' statement on line 3",
            (4, 1),
            (4, 5),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 11), reason="Syntax only permitted on 3.11+"
            ),
        ),
        (
            "try:\n\tpass\nexcept ValueError, IndexError as e:",
            SyntaxError,
            "multiple exception types must be parenthesized",
            (3, 8),
            (3, 35),
        ),
        pytest.param(
            "try:\n\tpass\nexcept* ValueError, IndexError as e:",
            SyntaxError,
            "multiple exception types must be parenthesized",
            (3, 9),
            (3, 36),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 11), reason="Syntax unsupported before 3.11+"
            ),
        ),
        (
            "try:\n\tpass\nexcept ValueError, IndexError:",
            SyntaxError,
            "multiple exception types must be parenthesized",
            (3, 8),
            (3, 30),
        ),
        (
            "try:\n\tpass\nexcept ValueError, IndexError,:",
            SyntaxError,
            "multiple exception types must be parenthesized",
            (3, 8),
            (3, 31),
        ),
        (
            "try:\n\tpass\nexcept ValueError, IndexError, a=1:",
            SyntaxError,
            "invalid syntax",
            (3, 18),
            (3, 19),
        ),
        (
            "try:\n\tpass\nexcept Exception\npass",
            SyntaxError,
            "expected ':'",
            (3, 17),
            (3, 18) if sys.version_info >= (3, 12) else (3, 17),
        ),
        (
            "try:\n\tpass\nexcept Exception as e\npass",
            SyntaxError,
            "expected ':'",
            (3, 22),
            (3, 23) if sys.version_info >= (3, 12) else (3, 22),
        ),
        (
            "try:\n\tpass\nexcept\npass",
            SyntaxError,
            "expected ':'",
            (3, 7),
            (3, 8) if sys.version_info >= (3, 12) else (3, 7),
        ),
        pytest.param(
            "try:\n\tpass\nexcept*:\n\tpass",
            SyntaxError,
            "expected one or more exception types",
            (3, 8),
            (3, 9),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 11), reason="Syntax unsupported before 3.11+"
            ),
        ),
        pytest.param(
            "try:\n\tpass\nexcept* ValueError as e:\n\tpass\nexcept:\n\tpass",
            SyntaxError,
            "cannot have both 'except' and 'except*'",
            (5, 1),
            (5, 7),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 11), reason="Syntax unsupported before 3.11+"
            ),
        ),
        pytest.param(
            "try:\n\tpass\nexcept ValueError as e:\n\tpass\nexcept* TypeError:\n\tpass",
            SyntaxError,
            "cannot have both 'except' and 'except*'",
            (5, 1),
            (5, 8),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 11), reason="Syntax unsupported before 3.11+"
            ),
        ),
    ],
)
def test_invalid_except_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, stop",
    [
        (
            "try:\n\tpass\nfinally:\npass",
            IndentationError,
            "expected an indented block after 'finally' statement on line 3",
            (4, 1),
            (4, 5),
        ),
        (
            "try:\n\tpass\nexcept Exception:\n\tpass\nfinally:\npass",
            IndentationError,
            "expected an indented block after 'finally' statement on line 5",
            (6, 1),
            (6, 5),
        ),
    ],
)
def test_invalid_finally_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, stop
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, stop
    )


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Valid only in Python 3.10+")
@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "match a\n\tpass",
            SyntaxError,
            "expected ':'",
            (1, 8),
            (1, 9) if sys.version_info >= (3, 12) else (1, 8),
        ),
        (
            "match a:\npass",
            IndentationError,
            "expected an indented block after 'match' statement on line 1",
            (2, 1),
            (2, 5),
        ),
    ],
)
def test_invalid_match_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Valid only in Python 3.10+")
@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "match a:\n\tcase 1\n\t\tpass",
            SyntaxError,
            "expected ':'",
            (2, 8),
            (2, 9) if sys.version_info >= (3, 12) else (2, 8),
        ),
        (
            "match a:\n\tcase 1:\n\tpass",
            IndentationError,
            "expected an indented block after 'case' statement on line 2",
            (3, 2),
            (3, 6),
        ),
    ],
)
def test_invalid_case_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.skipif(sys.version_info < (3, 10), reason="Valid only in Python 3.10+")
@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        # As pattern
        (
            "match a:\n\tcase 1 as _:\n\t\tpass",
            SyntaxError,
            "cannot use '_' as a target",
            (2, 12),
            (2, 13),
        ),
        (
            "match a:\n\tcase 1 as 1+1:\n\t\tpass",
            SyntaxError,
            "invalid pattern target",
            (2, 12),
            (2, 15),
        ),
        # Class pattern
        (
            "match a:\n\tcase Foo(z=1, y=2, x):\n\t\tpass",
            SyntaxError,
            "positional patterns follow keyword patterns",
            (2, 21),
            (2, 22),
        ),
        (
            "match a:\n\tcase Foo(a, z=1, y=2, x):\n\t\tpass",
            SyntaxError,
            "positional patterns follow keyword patterns",
            (2, 24),
            (2, 25),
        ),
        (
            "match a:\n\tcase Foo(z=1, x, y=2):\n\t\tpass",
            SyntaxError,
            "positional patterns follow keyword patterns",
            (2, 16),
            (2, 17),
        ),
        (
            "match a:\n\tcase Foo(a=b, c, d=e, f, g=h, i, j=k, ...):\n\t\tpass",
            SyntaxError,
            "positional patterns follow keyword patterns",
            (2, 16),
            (2, 17),
        ),
        (
            "match x:\n\tcase -1j + 1j:\n\t\tpass",
            SyntaxError,
            "real number required in complex literal",
            (2, 8),
            (2, 10),
        ),
        (
            "match x:\n\tcase -1 + 1:\n\t\tpass",
            SyntaxError,
            "imaginary number required in complex literal",
            (2, 12),
            (2, 13),
        ),
    ],
)
def test_invalid_case_pattern(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "if a\n\tpass",
            SyntaxError,
            "expected ':'",
            (1, 5),
            (1, 6) if sys.version_info >= (3, 12) else (1, 5),
        ),
        (
            "if a:\npass",
            IndentationError,
            "expected an indented block after 'if' statement on line 1",
            (2, 1),
            (2, 5),
        ),
    ],
)
def test_invalid_if_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "if a:\n\tpass\nelif a\n\tpass",
            SyntaxError,
            "expected ':'",
            (3, 7),
            (3, 8) if sys.version_info >= (3, 12) else (3, 7),
        ),
        (
            "if a:\n\tpass\nelif b:\npass",
            IndentationError,
            "expected an indented block after 'elif' statement on line 3",
            (4, 1),
            (4, 5),
        ),
    ],
)
def test_invalid_elif_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "if a:\n\tpass\nelse\n\tpass",
            SyntaxError,
            "expected ':'",
            (3, 5),
            (3, 6) if sys.version_info >= (3, 12) else (3, 5),
        ),
        (
            "if a:\n\tpass\nelse:\npass",
            IndentationError,
            "expected an indented block after 'else' statement on line 3",
            (4, 1),
            (4, 5),
        ),
    ],
)
def test_invalid_else_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "while a\n\tpass",
            SyntaxError,
            "expected ':'",
            (1, 8),
            (1, 9) if sys.version_info >= (3, 12) else (1, 8),
        ),
        (
            "while a:\npass",
            IndentationError,
            "expected an indented block after 'while' statement on line 1",
            (2, 1),
            (2, 5),
        ),
    ],
)
def test_invalid_while_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "for a in range(10)\npass",
            SyntaxError,
            "expected ':'",
            (1, 19),
            (1, 20) if sys.version_info >= (3, 12) else (1, 19),
        ),
        (
            "for a in range(10):\npass",
            IndentationError,
            "expected an indented block after 'for' statement on line 1",
            (2, 1),
            (2, 5),
        ),
        (
            "async for a in range(10):\npass",
            IndentationError,
            "expected an indented block after 'for' statement on line 1",
            (2, 1),
            (2, 5),
        ),
        ("for a in raise:\npass", SyntaxError, "invalid syntax", (1, 10), (1, 15)),
        (
            "async for a in raise:\npass",
            SyntaxError,
            "invalid syntax",
            (1, 16),
            (1, 21),
        ),
    ],
)
def test_invalid_for_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "def f():\npass",
            IndentationError,
            "expected an indented block after function definition on line 1",
            (2, 1),
            (2, 5),
        ),
        (
            "async def f():\npass",
            IndentationError,
            "expected an indented block after function definition on line 1",
            (2, 1),
            (2, 5),
        ),
        (
            "def f(a,):\npass",
            IndentationError,
            "expected an indented block after function definition on line 1",
            (2, 1),
            (2, 5),
        ),
        pytest.param(
            "def f[T]() -> None:\npass",
            IndentationError,
            "expected an indented block after function definition on line 1",
            (2, 1),
            (2, 5),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 12), reason="Syntax only permitted on 3.12+"
            ),
        ),
        (
            "def f() -> None:\npass",
            IndentationError,
            "expected an indented block after function definition on line 1",
            (2, 1),
            (2, 5),
        ),
        (
            "def f:",
            SyntaxError,
            "expected '('",
            (1, 6),
            (1, 7) if sys.version_info >= (3, 11) else (1, 6),
        ),
        (
            "async def f:",
            SyntaxError,
            "expected '('",
            (1, 12),
            (1, 13) if sys.version_info >= (3, 11) else (1, 12),
        ),
    ],
)
def test_invalid_def_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file,
        python_parse_str,
        tmp_path,
        source,
        exception,
        message,
        start,
        end,
        (3, 11) if exception is SyntaxError else (3, 13),
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "class A\npass",
            SyntaxError,
            "expected ':'",
            (1, 8),
            (1, 9) if sys.version_info >= (3, 12) else (1, 8),
        ),
        (
            "class A:\npass",
            IndentationError,
            "expected an indented block after class definition on line 1",
            (2, 1),
            (2, 5),
        ),
        (
            "class A(object):\npass",
            IndentationError,
            "expected an indented block after class definition on line 1",
            (2, 1),
            (2, 5),
        ),
        pytest.param(
            "class A[T]:\npass",
            IndentationError,
            "expected an indented block after class definition on line 1",
            (2, 1),
            (2, 5),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 12), reason="Syntax only permitted on 3.12+"
            ),
        ),
        pytest.param(
            "class A[T](object):\npass",
            IndentationError,
            "expected an indented block after class definition on line 1",
            (2, 1),
            (2, 5),
            marks=pytest.mark.skipif(
                sys.version_info < (3, 12), reason="Syntax only permitted on 3.12+"
            ),
        ),
    ],
)
def test_invalid_class_stmt(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file,
        python_parse_str,
        tmp_path,
        source,
        exception,
        message,
        start,
        end,
        (3, 13),
    )


@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        # NOTE CPython reports 0 as the end offset which looks odd to me
        ("{a: 1, b}", SyntaxError, "':' expected after dictionary key", (1, 8), (1, None)),
        ("{a: 1, c: 2, b}", SyntaxError, "':' expected after dictionary key", (1, 14), (1, None)),
        (
            "{a: 1, b:}",
            SyntaxError,
            "expression expected after dictionary key and ':'",
            (1, 9),
            (1, 10),
        ),
        (
            "{c: 1, a: *b}",
            SyntaxError,
            "cannot use a starred expression in a dictionary value",
            (1, 11),
            (1, 13),
        ),
        ("{b:}", SyntaxError, "expression expected after dictionary key and ':'", (1, 3), (1, 4)),
        (
            "{b:, c}",
            SyntaxError,
            "expression expected after dictionary key and ':'",
            (1, 3),
            (1, 4),
        ),
        (
            "{a: *b}",
            SyntaxError,
            "cannot use a starred expression in a dictionary value",
            (1, 5),
            (1, 7),
        ),
        (
            "{**c, a: *b}",
            SyntaxError,
            "cannot use a starred expression in a dictionary value",
            (1, 10),
            (1, 12),
        ),
    ],
)
def test_invalid_dict_key_value(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )


@pytest.mark.skipif(sys.version_info < (3, 12), reason="Requires Python 3.12+")
@pytest.mark.parametrize(
    "source, exception, message, start, end",
    [
        (
            "def f[*T: (int, float)]():\n\tpass",
            SyntaxError,
            "cannot use constraints with TypeVarTuple",
            (1, 9),
            (1, 23),
        ),
        (
            "def f[*T: int]():\n\tpass",
            SyntaxError,
            "cannot use bound with TypeVarTuple",
            (1, 9),
            (1, 14),
        ),
        (
            "def f[**T: (int,)]():\n\tpass",
            SyntaxError,
            "cannot use constraints with ParamSpec",
            (1, 10),
            (1, 18),
        ),
        (
            "def f[**T: int]():\n\tpass",
            SyntaxError,
            "cannot use bound with ParamSpec",
            (1, 10),
            (1, 15),
        ),
    ],
)
def test_syntax_error_in_type_params(
    python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
):
    parse_invalid_syntax(
        python_parse_file, python_parse_str, tmp_path, source, exception, message, start, end
    )



"""Test pure Python parser against cpython parser."""
import ast
import difflib
import io
import sys
import textwrap
import tokenize
from pathlib import Path

import pytest


@pytest.mark.parametrize(
    "filename",
    [
        pytest.param(
            "advanced_decorators.py",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 9), reason="Valid only in Python 3.9+"
            ),
        ),
        "assignment.py",
        "async.py",
        "call.py",
        "comprehensions.py",
        "expressions.py",
        "fstrings.py",
        "function_def.py",
        "imports.py",
        "lambdas.py",
        pytest.param(
            "multi_statement_per_line.py",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 9), reason="Col offset match only on Python 3.9+"
            ),
        ),
        "no_newline_at_end_of_file.py",
        "no_newline_at_end_of_file_with_comment.py",
        pytest.param(
            "pattern_matching.py",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 10), reason="Valid only in Python 3.10+"
            ),
        ),
        "simple_decorators.py",
        "statements.py",
        pytest.param(
            "try_except_group.py",
            marks=pytest.mark.skipif(
                sys.version_info <= (3, 11), reason="except* allowed only in Python 3.11+"
            ),
        ),
        pytest.param(
            "type_params.py",
            marks=pytest.mark.skipif(
                sys.version_info <= (3, 12),
                reason="type declarations allowed only in Python 3.12+",
            ),
        ),
        pytest.param(
            "with_statement_multi_items.py",
            marks=pytest.mark.skipif(
                sys.version_info < (3, 9),
                reason="Parenthesized with items allowed only in Python 3.9+",
            ),
        ),
    ],
)
def test_parser(python_parse_file, python_parse_str, filename):
    path = Path(__file__).parent / "data" / filename
    with open(path) as f:
        source = f.read()

    for part in source.split("\n\n\n"):
        original = ast.parse(part)

        kwargs = dict(include_attributes=True)
        if sys.version_info >= (3, 9):
            kwargs["indent"] = "  "

        try:
            pp_ast = python_parse_str(part, "exec")
        except Exception:
            temp = io.StringIO(part)
            print("Parsing failed:")
            print("Source is:")
            print(textwrap.indent(part, "  "))
            temp = io.StringIO(part)
            print("Token stream is:")
            for t in tokenize.generate_tokens(temp.readline):
                print(t)
            print()
            print("CPython ast is:")
            print(ast.dump(original, **kwargs))
            raise

        o = ast.dump(original, **kwargs)
        p = ast.dump(pp_ast, **kwargs)
        diff = "\n".join(
            difflib.unified_diff(o.split("\n"), p.split("\n"), "cpython", "python-pegen")
        )
        if diff:
            print(part)
            print(diff)
        assert not diff

    o = ast.dump(ast.parse(source), **kwargs)
    p = ast.dump(python_parse_file(path), **kwargs)
    diff = "\n".join(difflib.unified_diff(o.split("\n"), p.split("\n"), "cpython", "python-pegen"))
    assert not diff



a + b
a - b
a * b
a / b
a // b
a % b
a @ b
a << b
a >> b
a | b
a ^ b
a ** b
a == b
a < b
a <= b
a > b
a >= b
a != b
a & b
~a
(1, 2, 3)
["a", "b"]
{1, 2}
{a: a.b}
{**d, a: b}

not b
a if b else c
a or b
a and b
a in b
a not in b
a is b
a is not b

a * (+1)
a * (-1)
a * (~1)

(a)
(yield a)


def f():
    pass


def f() -> None:
    pass


def f(a):
    pass


def f(a: int) -> Tuple[int, ...]:
    pass


def f(a: int = 1) -> Tuple[int, ...]:
    pass


def f(a, b: int):
    pass


def f(a: bool, b: int = 1):
    pass


def f(a, /):
    pass


def f(a=1, /):
    pass


def f(a, b=1, /):
    pass


def f(a, /, b):
    pass


def f(a, c=2, /, b=5):
    pass


def f(a, /, b=1):
    pass


def f(a, *, b):
    pass


def f(a, *, b, c=1):
    pass


def f(a, *, b=1):
    pass


def f(*, b):
    pass


def f(*, b, c=1):
    pass


def f(*, b=1):
    pass


def f(b=1, *c):
    pass


def f(*args):
    pass


def f(**kwargs):
    pass


def f(a, **kwargs):
    pass


def f(a=1, **kwargs):
    pass


def f(*, a=1, **kwargs):
    pass


def f(*a, **b):
    pass


def f(a, /, b, *, v=1, **d):
    pass


async def f():
    pass


async def f() -> None:
    pass


async def f(a):
    pass


async def f(a: int) -> Tuple[int, ...]:
    pass


async def f(a: int = 1) -> Tuple[int, ...]:
    pass


async def f(a, b: int):
    pass


async def f(a: bool, b: int = 1):
    pass


async def f(a, /):
    pass


async def f(a=1, /):
    pass


async def f(a, b=1, /):
    pass


async def f(a, /, b):
    pass


async def f(a, c=2, /, b=5):
    pass


async def f(a, /, b=1):
    pass


async def f(a, *, b):
    pass


async def f(a, *, b=1):
    pass


async def f(*, b):
    pass


async def f(*, b=1):
    pass


async def f(b=1, *c):
    pass


async def f(*args):
    pass


async def f(**kwargs):
    pass


async def f(a, **kwargs):
    pass


async def f(a=1, **kwargs):
    pass


async def f(*, a=1, **kwargs):
    pass


async def f(*a, **b):
    pass


async def f(a, /, b, *, v=1, **d):
    pass



type TA1 = int
type TA2 = TA1 | str



type NonGeneric = int
type Generic[A] = dict[A, A]
type VeryGeneric[T, *Ts, **P] = Callable[P, tuple[T, *Ts]]




def outer[A]():
    type TA1[B] = dict[A, B]
    return TA1



class Parent[A]:
    type TA1[B] = dict[A, B]




class Outer[A]:
    def inner[B](self):
        type TA1[C] = TA1[A, B] | int
        return TA1



def more_generic[T, *Ts, **P]():
    type TA[T2, *Ts2, **P2] = tuple[Callable[P, tuple[T, *Ts]], Callable[P2, tuple[T2, *Ts2]]]
    return TA



a = 10
f'{a * x()}'



f'no formatted values'
f'eggs {a * x()} spam {b + y()}'



a = 10
f'{a * x()} {a * x()} {a * x()}'



a = 10
f'''
  {a
     *
       x()}
non-important content
'''



a = f'''
          {blech}
    '''



x = (
    f" {test(t)}"
)



x = (
    u'wat',
    u"wat",
    b'wat',
    b"wat",
    f'wat',
    f"wat",
)
y = (
    u'''wat''',
    u"""wat""",
    b'''wat''',
    b"""wat""",
    f'''wat''',
    f"""wat""",
)



x = (
        'PERL_MM_OPT', (
            f'wat'
            f'some_string={f(x)} '
            f'wat'
        ),
)



f'{expr:}'
f'{expr:d}'
foo = 3.14159
verbosePrint(f'Foo {foo:.3} bar.')



import ast
import difflib
import io
import textwrap
from tokenize import NAME, NEWLINE, NUMBER, OP, TokenInfo
from typing import Any, Dict, Type

import pytest
from pegen.grammar import Grammar, GrammarError
from pegen.grammar_parser import GeneratedParser as GrammarParser
from pegen.parser import Parser
from pegen.python_generator import PythonParserGenerator
from pegen.utils import generate_parser, make_parser, parse_string


def test_parse_grammar() -> None:
    grammar_source = """
    start: sum NEWLINE
    sum: t1=term '+' t2=term { action } | term
    term: NUMBER
    """
    expected = """
    start: sum NEWLINE
    sum: term '+' term | term
    term: NUMBER
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    rules = grammar.rules
    assert str(grammar) == textwrap.dedent(expected).strip()
    # Check the str() and repr() of a few rules; AST nodes don't support ==.
    assert str(rules["start"]) == "start: sum NEWLINE"
    assert str(rules["sum"]) == "sum: term '+' term | term"
    expected_repr = "Rule('term', None, Rhs([Alt([NamedItem(None, NameLeaf('NUMBER'))])]))"
    assert repr(rules["term"]) == expected_repr


def test_parse_grammar_with_types() -> None:
    grammar = """
    start[ast.BinOp]: term ('+' term)* NEWLINE
    term[T[int]]: NUMBER
    c_rule[expr_ty*]: a=NUMBER? { _new_expr_ty(a) }
    """

    grammar: Grammar = parse_string(grammar, GrammarParser)
    rules = grammar.rules
    assert rules["start"].type.replace(" ", "") == "ast.BinOp"
    assert rules["term"].type.replace(" ", "") == "T[int]"
    assert rules["c_rule"].type == "expr_ty*"


def test_long_rule_str() -> None:
    grammar_source = """
    start: zero | one | one zero | one one | one zero zero | one zero one | one one zero | one one one
    """
    expected = """
    start:
        | zero
        | one
        | one zero
        | one one
        | one zero zero
        | one zero one
        | one one zero
        | one one one
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    assert str(grammar.rules["start"]) == textwrap.dedent(expected).strip()


def test_typed_rules() -> None:
    grammar = """
    start[int]: sum NEWLINE
    sum[int]: t1=term '+' t2=term { action } | term
    term[int]: NUMBER
    """
    rules = parse_string(grammar, GrammarParser).rules
    # Check the str() and repr() of a few rules; AST nodes don't support ==.
    assert str(rules["start"]) == "start: sum NEWLINE"
    assert str(rules["sum"]) == "sum: term '+' term | term"
    assert (
        repr(rules["term"])
        == "Rule('term', 'int', Rhs([Alt([NamedItem(None, NameLeaf('NUMBER'))])]))"
    )


def test_gather() -> None:
    grammar = """
    start: ','.thing+ NEWLINE
    thing: NUMBER
    """
    rules = parse_string(grammar, GrammarParser).rules
    assert str(rules["start"]) == "start: ','.thing+ NEWLINE"
    print(repr(rules["start"]))
    assert repr(rules["start"]).startswith(
        "Rule('start', None, Rhs([Alt([NamedItem(None, Gather(StringLeaf(\"','\"), NameLeaf('thing'"
    )
    assert str(rules["thing"]) == "thing: NUMBER"
    parser_class = make_parser(grammar)
    node = parse_string("42\n", parser_class)
    assert node == [
        [TokenInfo(NUMBER, string="42", start=(1, 0), end=(1, 2), line="42\n")],
        TokenInfo(NEWLINE, string="\n", start=(1, 2), end=(1, 3), line="42\n"),
    ]
    node = parse_string("1, 2\n", parser_class)
    assert node == [
        [
            TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1, 2\n"),
            TokenInfo(NUMBER, string="2", start=(1, 3), end=(1, 4), line="1, 2\n"),
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 4), end=(1, 5), line="1, 2\n"),
    ]


def test_expr_grammar() -> None:
    grammar = """
    start: sum NEWLINE
    sum: term '+' term | term
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("42\n", parser_class)
    assert node == [
        TokenInfo(NUMBER, string="42", start=(1, 0), end=(1, 2), line="42\n"),
        TokenInfo(NEWLINE, string="\n", start=(1, 2), end=(1, 3), line="42\n"),
    ]


def test_optional_operator() -> None:
    grammar = """
    start: sum NEWLINE
    sum: term ('+' term)?
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1+2\n", parser_class)
    assert node == [
        [
            TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1+2\n"),
            [
                TokenInfo(OP, string="+", start=(1, 1), end=(1, 2), line="1+2\n"),
                TokenInfo(NUMBER, string="2", start=(1, 2), end=(1, 3), line="1+2\n"),
            ],
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 3), end=(1, 4), line="1+2\n"),
    ]
    node = parse_string("1\n", parser_class)
    assert node == [
        [TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1\n"), None],
        TokenInfo(NEWLINE, string="\n", start=(1, 1), end=(1, 2), line="1\n"),
    ]


def test_optional_literal() -> None:
    grammar = """
    start: sum NEWLINE
    sum: term '+' ?
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1+\n", parser_class)
    assert node == [
        [
            TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1+\n"),
            TokenInfo(OP, string="+", start=(1, 1), end=(1, 2), line="1+\n"),
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 2), end=(1, 3), line="1+\n"),
    ]
    node = parse_string("1\n", parser_class)
    assert node == [
        [TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1\n"), None],
        TokenInfo(NEWLINE, string="\n", start=(1, 1), end=(1, 2), line="1\n"),
    ]


def test_alt_optional_operator() -> None:
    grammar = """
    start: sum NEWLINE
    sum: term ['+' term]
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1 + 2\n", parser_class)
    assert node == [
        [
            TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1 + 2\n"),
            [
                TokenInfo(OP, string="+", start=(1, 2), end=(1, 3), line="1 + 2\n"),
                TokenInfo(NUMBER, string="2", start=(1, 4), end=(1, 5), line="1 + 2\n"),
            ],
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 5), end=(1, 6), line="1 + 2\n"),
    ]
    node = parse_string("1\n", parser_class)
    assert node == [
        [TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1\n"), None],
        TokenInfo(NEWLINE, string="\n", start=(1, 1), end=(1, 2), line="1\n"),
    ]


def test_repeat_0_simple() -> None:
    grammar = """
    start: thing thing* NEWLINE
    thing: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1 2 3\n", parser_class)
    assert node == [
        TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1 2 3\n"),
        [
            TokenInfo(NUMBER, string="2", start=(1, 2), end=(1, 3), line="1 2 3\n"),
            TokenInfo(NUMBER, string="3", start=(1, 4), end=(1, 5), line="1 2 3\n"),
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 5), end=(1, 6), line="1 2 3\n"),
    ]
    node = parse_string("1\n", parser_class)
    assert node == [
        TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1\n"),
        [],
        TokenInfo(NEWLINE, string="\n", start=(1, 1), end=(1, 2), line="1\n"),
    ]


def test_repeat_0_complex() -> None:
    grammar = """
    start: term ('+' term)* NEWLINE
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1 + 2 + 3\n", parser_class)
    assert node == [
        TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1 + 2 + 3\n"),
        [
            [
                TokenInfo(OP, string="+", start=(1, 2), end=(1, 3), line="1 + 2 + 3\n"),
                TokenInfo(NUMBER, string="2", start=(1, 4), end=(1, 5), line="1 + 2 + 3\n"),
            ],
            [
                TokenInfo(OP, string="+", start=(1, 6), end=(1, 7), line="1 + 2 + 3\n"),
                TokenInfo(NUMBER, string="3", start=(1, 8), end=(1, 9), line="1 + 2 + 3\n"),
            ],
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 9), end=(1, 10), line="1 + 2 + 3\n"),
    ]


def test_repeat_1_simple() -> None:
    grammar = """
    start: thing thing+ NEWLINE
    thing: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1 2 3\n", parser_class)
    assert node == [
        TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1 2 3\n"),
        [
            TokenInfo(NUMBER, string="2", start=(1, 2), end=(1, 3), line="1 2 3\n"),
            TokenInfo(NUMBER, string="3", start=(1, 4), end=(1, 5), line="1 2 3\n"),
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 5), end=(1, 6), line="1 2 3\n"),
    ]
    with pytest.raises(SyntaxError):
        parse_string("1\n", parser_class)


def test_repeat_1_complex() -> None:
    grammar = """
    start: term ('+' term)+ NEWLINE
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1 + 2 + 3\n", parser_class)
    assert node == [
        TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1 + 2 + 3\n"),
        [
            [
                TokenInfo(OP, string="+", start=(1, 2), end=(1, 3), line="1 + 2 + 3\n"),
                TokenInfo(NUMBER, string="2", start=(1, 4), end=(1, 5), line="1 + 2 + 3\n"),
            ],
            [
                TokenInfo(OP, string="+", start=(1, 6), end=(1, 7), line="1 + 2 + 3\n"),
                TokenInfo(NUMBER, string="3", start=(1, 8), end=(1, 9), line="1 + 2 + 3\n"),
            ],
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 9), end=(1, 10), line="1 + 2 + 3\n"),
    ]
    with pytest.raises(SyntaxError):
        parse_string("1\n", parser_class)


def test_repeat_with_sep_simple() -> None:
    grammar = """
    start: ','.thing+ NEWLINE
    thing: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("1, 2, 3\n", parser_class)
    assert node == [
        [
            TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1, 2, 3\n"),
            TokenInfo(NUMBER, string="2", start=(1, 3), end=(1, 4), line="1, 2, 3\n"),
            TokenInfo(NUMBER, string="3", start=(1, 6), end=(1, 7), line="1, 2, 3\n"),
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 7), end=(1, 8), line="1, 2, 3\n"),
    ]


def test_left_recursive() -> None:
    grammar_source = """
    start: expr NEWLINE
    expr: ('-' term | expr '+' term | term)
    term: NUMBER
    foo: NAME+
    bar: NAME*
    baz: NAME?
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    parser_class = generate_parser(grammar)
    rules = grammar.rules
    assert not rules["start"].left_recursive
    assert rules["expr"].left_recursive
    assert not rules["term"].left_recursive
    assert not rules["foo"].left_recursive
    assert not rules["bar"].left_recursive
    assert not rules["baz"].left_recursive
    node = parse_string("1 + 2 + 3\n", parser_class)
    assert node == [
        [
            [
                TokenInfo(NUMBER, string="1", start=(1, 0), end=(1, 1), line="1 + 2 + 3\n"),
                TokenInfo(OP, string="+", start=(1, 2), end=(1, 3), line="1 + 2 + 3\n"),
                TokenInfo(NUMBER, string="2", start=(1, 4), end=(1, 5), line="1 + 2 + 3\n"),
            ],
            TokenInfo(OP, string="+", start=(1, 6), end=(1, 7), line="1 + 2 + 3\n"),
            TokenInfo(NUMBER, string="3", start=(1, 8), end=(1, 9), line="1 + 2 + 3\n"),
        ],
        TokenInfo(NEWLINE, string="\n", start=(1, 9), end=(1, 10), line="1 + 2 + 3\n"),
    ]


def test_python_expr() -> None:
    grammar = """
    start: expr NEWLINE? $ { ast.Expression(expr, LOCATIONS) }
    expr: ( expr '+' term { ast.BinOp(expr, ast.Add(), term, LOCATIONS) }
          | expr '-' term { ast.BinOp(expr, ast.Sub(), term, LOCATIONS) }
          | term { term }
          )
    term: ( l=term '*' r=factor { ast.BinOp(l, ast.Mult(), r, LOCATIONS) }
          | l=term '/' r=factor { ast.BinOp(l, ast.Div(), r, LOCATIONS) }
          | factor { factor }
          )
    factor: ( '(' expr ')' { expr }
            | atom { atom }
            )
    atom: ( n=NAME { ast.Name(id=n.string, ctx=ast.Load(), LOCATIONS) }
          | n=NUMBER { ast.Constant(value=ast.literal_eval(n.string), LOCATIONS) }
          )
    """
    parser_class = make_parser(grammar)
    node = parse_string("(1 + 2*3 + 5)/(6 - 2)\n", parser_class)
    code = compile(node, "", "eval")
    val = eval(code)
    assert val == 3.0


def test_nullable() -> None:
    grammar_source = """
    start: sign NUMBER
    sign: ['-' | '+']
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    out = io.StringIO()
    # Called to visit the grammar which updates the rules
    PythonParserGenerator(grammar, out)
    rules = grammar.rules
    assert rules["start"].nullable is False  # Not None!
    assert rules["sign"].nullable


def test_advanced_left_recursive() -> None:
    grammar_source = """
    start: NUMBER | sign start
    sign: ['-']
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    out = io.StringIO()
    # Called to visit the grammar which updates the rules
    PythonParserGenerator(grammar, out)
    rules = grammar.rules
    assert rules["start"].nullable is False  # Not None!
    assert rules["sign"].nullable
    assert rules["start"].left_recursive
    assert not rules["sign"].left_recursive


def test_mutually_left_recursive() -> None:
    grammar_source = """
    start: foo 'E'
    foo: bar 'A' | 'B'
    bar: foo 'C' | 'D'
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    out = io.StringIO()
    genr = PythonParserGenerator(grammar, out)
    rules = grammar.rules
    assert not rules["start"].left_recursive
    assert rules["foo"].left_recursive
    assert rules["bar"].left_recursive
    genr.generate("<string>")
    ns: Dict[str, Any] = {}
    exec(out.getvalue(), ns)
    parser_class: Type[Parser] = ns["GeneratedParser"]
    node = parse_string("D A C A E", parser_class)
    assert node == [
        [
            [
                [
                    TokenInfo(type=NAME, string="D", start=(1, 0), end=(1, 1), line="D A C A E"),
                    TokenInfo(type=NAME, string="A", start=(1, 2), end=(1, 3), line="D A C A E"),
                ],
                TokenInfo(type=NAME, string="C", start=(1, 4), end=(1, 5), line="D A C A E"),
            ],
            TokenInfo(type=NAME, string="A", start=(1, 6), end=(1, 7), line="D A C A E"),
        ],
        TokenInfo(type=NAME, string="E", start=(1, 8), end=(1, 9), line="D A C A E"),
    ]
    node = parse_string("B C A E", parser_class)
    assert node is not None
    assert node == [
        [
            [
                TokenInfo(type=NAME, string="B", start=(1, 0), end=(1, 1), line="B C A E"),
                TokenInfo(type=NAME, string="C", start=(1, 2), end=(1, 3), line="B C A E"),
            ],
            TokenInfo(type=NAME, string="A", start=(1, 4), end=(1, 5), line="B C A E"),
        ],
        TokenInfo(type=NAME, string="E", start=(1, 6), end=(1, 7), line="B C A E"),
    ]


def test_nasty_mutually_left_recursive() -> None:
    # This grammar does not recognize 'x - + =', much to my chagrin.
    # But that's the way PEG works.
    # [Breathlessly]
    # The problem is that the toplevel target call
    # recurses into maybe, which recognizes 'x - +',
    # and then the toplevel target looks for another '+',
    # which fails, so it retreats to NAME,
    # which succeeds, so we end up just recognizing 'x',
    # and then start fails because there's no '=' after that.
    grammar_source = """
    start: target '='
    target: maybe '+' | NAME
    maybe: maybe '-' | target
    """
    grammar: Grammar = parse_string(grammar_source, GrammarParser)
    out = io.StringIO()
    genr = PythonParserGenerator(grammar, out)
    genr.generate("<string>")
    ns: Dict[str, Any] = {}
    exec(out.getvalue(), ns)
    parser_class = ns["GeneratedParser"]
    with pytest.raises(SyntaxError):
        parse_string("x - + =", parser_class)


def test_lookahead() -> None:
    grammar = """
    start: (expr_stmt | assign_stmt) &'.'
    expr_stmt: !(target '=') expr
    assign_stmt: target '=' expr
    expr: term ('+' term)*
    target: NAME
    term: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("foo = 12 + 12 .", parser_class)
    assert node == [
        TokenInfo(NAME, string="foo", start=(1, 0), end=(1, 3), line="foo = 12 + 12 ."),
        TokenInfo(OP, string="=", start=(1, 4), end=(1, 5), line="foo = 12 + 12 ."),
        [
            TokenInfo(NUMBER, string="12", start=(1, 6), end=(1, 8), line="foo = 12 + 12 ."),
            [
                [
                    TokenInfo(OP, string="+", start=(1, 9), end=(1, 10), line="foo = 12 + 12 ."),
                    TokenInfo(
                        NUMBER, string="12", start=(1, 11), end=(1, 13), line="foo = 12 + 12 ."
                    ),
                ]
            ],
        ],
    ]


def test_named_lookahead_error() -> None:
    grammar = """
    start: foo=!'x' NAME
    """
    with pytest.raises(SyntaxError):
        make_parser(grammar)


def test_start_leader() -> None:
    grammar = """
    start: attr | NAME
    attr: start '.' NAME
    """
    # Would assert False without a special case in compute_left_recursives().
    make_parser(grammar)


def test_left_recursion_too_complex() -> None:
    grammar = """
    start: foo
    foo: bar '+' | baz '+' | '+'
    bar: baz '-' | foo '-' | '-'
    baz: foo '*' | bar '*' | '*'
    """
    with pytest.raises(ValueError) as errinfo:
        make_parser(grammar)
    assert "no leader" in str(errinfo.value)


def test_cut() -> None:
    grammar = """
    start: '(' ~ expr ')'
    expr: NUMBER
    """
    parser_class = make_parser(grammar)
    node = parse_string("(1)", parser_class, verbose=True)
    assert node == [
        TokenInfo(OP, string="(", start=(1, 0), end=(1, 1), line="(1)"),
        TokenInfo(NUMBER, string="1", start=(1, 1), end=(1, 2), line="(1)"),
        TokenInfo(OP, string=")", start=(1, 2), end=(1, 3), line="(1)"),
    ]


def test_cut_early_exit() -> None:
    grammar = """
    start: '(' ~ expr ')' | '(' name ')'
    expr: NUMBER
    name: NAME
    """
    parser_class = make_parser(grammar)
    with pytest.raises(SyntaxError):
        parse_string("(a)", parser_class, verbose=True)


def test_dangling_reference() -> None:
    grammar = """
    start: foo ENDMARKER
    foo: bar NAME
    """
    with pytest.raises(GrammarError):
        make_parser(grammar)


def test_bad_token_reference() -> None:
    grammar = """
    start: foo
    foo: NAMEE
    """
    with pytest.raises(GrammarError):
        make_parser(grammar)


def test_missing_start() -> None:
    grammar = """
    foo: NAME
    """
    with pytest.raises(GrammarError):
        make_parser(grammar)


def test_soft_keyword() -> None:
    grammar = """
    start:
        | "number" n=NUMBER { eval(n.string) }
        | "string" n=STRING { n.string }
        | SOFT_KEYWORD l=NAME n=(NUMBER | NAME | STRING) { f"{l.string} = {n.string}"}
    """
    parser_class = make_parser(grammar)
    assert parse_string("number 1", parser_class, verbose=True) == 1
    assert parse_string("string 'b'", parser_class, verbose=True) == "'b'"
    assert parse_string("number test 1", parser_class, verbose=True) == "test = 1"
    assert parse_string("string test 'b'", parser_class, verbose=True) == "test = 'b'"
    with pytest.raises(SyntaxError):
        parse_string("test 1", parser_class, verbose=True)


def test_forced() -> None:
    grammar = """
    start: NAME &&':' | NAME
    """
    parser_class = make_parser(grammar)
    assert parse_string("number :", parser_class, verbose=True)
    with pytest.raises(SyntaxError) as e:
        parse_string("a", parser_class, verbose=True)

    assert "expected ':'" in str(e.exconly())


def test_forced_with_group() -> None:
    grammar = """
    start: NAME &&(':' | ';') | NAME
    """
    parser_class = make_parser(grammar)
    assert parse_string("number :", parser_class, verbose=True)
    assert parse_string("number ;", parser_class, verbose=True)
    with pytest.raises(SyntaxError) as e:
        parse_string("a", parser_class, verbose=True)

    assert "expected (':' | ';')" in e.value.args[0]


UNREACHABLE = "None  # This is a test"


def test_unreachable_explicit() -> None:
    source = """
    start: NAME { UNREACHABLE }
    """
    grammar = parse_string(source, GrammarParser)
    out = io.StringIO()
    genr = PythonParserGenerator(grammar, out, unreachable_formatting=UNREACHABLE)
    genr.generate("<string>")
    assert UNREACHABLE in out.getvalue()


def test_unreachable_implicit1() -> None:
    source = """
    start: NAME | invalid_input
    invalid_input: NUMBER { None }
    """
    grammar = parse_string(source, GrammarParser)
    out = io.StringIO()
    genr = PythonParserGenerator(grammar, out, unreachable_formatting=UNREACHABLE)
    genr.generate("<string>")
    assert UNREACHABLE in out.getvalue()


def test_unreachable_implicit2() -> None:
    source = """
    start: NAME | '(' invalid_input ')'
    invalid_input: NUMBER { None }
    """
    grammar = parse_string(source, GrammarParser)
    out = io.StringIO()
    genr = PythonParserGenerator(grammar, out, unreachable_formatting=UNREACHABLE)
    genr.generate("<string>")
    assert UNREACHABLE in out.getvalue()


def test_unreachable_implicit3() -> None:
    source = """
    start: NAME | invalid_input { None }
    invalid_input: NUMBER
    """
    grammar = parse_string(source, GrammarParser)
    out = io.StringIO()
    genr = PythonParserGenerator(grammar, out, unreachable_formatting=UNREACHABLE)
    genr.generate("<string>")
    assert UNREACHABLE not in out.getvalue()


def test_locations_in_alt_action_and_group() -> None:
    grammar = """
    start: t=term NEWLINE? $ { ast.Expression(t, LOCATIONS) }
    term:
        | l=term '*' r=factor { ast.BinOp(l, ast.Mult(), r, LOCATIONS) }
        | l=term '/' r=factor { ast.BinOp(l, ast.Div(), r, LOCATIONS) }
        | factor
    factor:
        | (
            n=NAME { ast.Name(id=n.string, ctx=ast.Load(), LOCATIONS) } |
            n=NUMBER { ast.Constant(value=ast.literal_eval(n.string), LOCATIONS) }
         )
    """
    parser_class = make_parser(grammar)
    source = "2*3\n"
    o = ast.dump(parse_string(source, parser_class).body, include_attributes=True)
    p = ast.dump(ast.parse(source).body[0].value, include_attributes=True).replace(
        " kind=None,", ""
    )
    diff = "\n".join(difflib.unified_diff(o.split("\n"), p.split("\n"), "cpython", "python-pegen"))
    if diff:
        print(diff)
    assert not diff


def test_keywords() -> None:
    grammar = """
    start: 'one' 'two' 'three' 'four' 'five' "six" "seven" "eight" "nine" "ten"
    """
    parser_class = make_parser(grammar)
    assert parser_class.KEYWORDS == ("five", "four", "one", "three", "two")
    assert parser_class.SOFT_KEYWORDS == ("eight", "nine", "seven", "six", "ten")



#!/usr/bin/env python3.8

"""pegen -- PEG Generator.

Search the web for PEG Parsers for reference.
"""

import argparse
import sys
import time
import token
import traceback
from typing import Tuple

from pegen.build import (
    Grammar,
    Parser,
    ParserGenerator,
    Tokenizer,
    build_python_parser_and_generator,
)
from pegen.validator import validate_grammar


def generate_python_code(
    args: argparse.Namespace,
) -> Tuple[Grammar, Parser, Tokenizer, ParserGenerator]:
    verbose = args.verbose
    verbose_tokenizer = verbose >= 3
    verbose_parser = verbose == 2 or verbose >= 4
    try:
        grammar, parser, tokenizer, gen = build_python_parser_and_generator(
            args.grammar_filename,
            args.output,
            verbose_tokenizer,
            verbose_parser,
            skip_actions=args.skip_actions,
        )
        return grammar, parser, tokenizer, gen
    except Exception as err:
        if args.verbose:
            raise  # Show traceback
        traceback.print_exception(err.__class__, err, None)
        sys.stderr.write("For full traceback, use -v\n")
        sys.exit(1)


argparser = argparse.ArgumentParser(
    prog="pegen", description="Experimental PEG-like parser generator"
)
argparser.add_argument("-q", "--quiet", action="store_true", help="Don't print the parsed grammar")
argparser.add_argument(
    "-v",
    "--verbose",
    action="count",
    default=0,
    help="Print timing stats; repeat for more debug output",
)

argparser.add_argument("grammar_filename", help="Grammar description")
argparser.add_argument(
    "-o",
    "--output",
    metavar="OUT",
    default="parse.py",
    help="Where to write the generated parser",
)
argparser.add_argument(
    "--skip-actions",
    action="store_true",
    help="Suppress code emission for rule actions",
)


def main() -> None:
    args = argparser.parse_args()

    t0 = time.time()
    grammar, parser, tokenizer, gen = generate_python_code(args)
    t1 = time.time()

    validate_grammar(grammar)

    if not args.quiet:
        if args.verbose:
            print("Raw Grammar:")
            for line in repr(grammar).splitlines():
                print(" ", line)

        print("Clean Grammar:")
        for line in str(grammar).splitlines():
            print(" ", line)

    if args.verbose:
        print("First Graph:")
        for src, dsts in gen.first_graph.items():
            print(f"  {src} -> {', '.join(dsts)}")
        print("First SCCS:")
        for scc in gen.first_sccs:
            print(" ", scc, end="")
            if len(scc) > 1:
                print(
                    "  # Indirectly left-recursive; leaders:",
                    {name for name in scc if grammar.rules[name].leader},
                )
            else:
                name = next(iter(scc))
                if name in gen.first_graph[name]:
                    print("  # Left-recursive")
                else:
                    print()

    if args.verbose:
        dt = t1 - t0
        diag = tokenizer.diagnose()
        nlines = diag.end[0]
        if diag.type == token.ENDMARKER:
            nlines -= 1
        print(f"Total time: {dt:.3f} sec; {nlines} lines", end="")
        if dt:
            print(f"; {nlines / dt:.0f} lines/sec")
        else:
            print()
        print("Caches sizes:")
        print(f"  token array : {len(tokenizer._tokens):10}")
        print(f"        cache : {len(parser._cache):10}")


if __name__ == "__main__":
    main()



lambda: 1

lambda x: x

lambda x,: x

lambda x=1: x

lambda x, y: x + y

lambda x, /: x

lambda x, y=1, /: x + y

lambda x, /, y: x + y

lambda x, /, y=1, z=2: x + y + z

lambda x, y=1, /, z=5: x + y + z

lambda x=1, /, *y: x + y

lambda x, *, y: x + y

lambda x, *, y, z: x + y + z

lambda *, x: x

lambda *x: x

lambda **x: x

lambda x, **y: y



class A:
    pass


class A(B):
    pass


class A(
    B,
    C,
):
    pass


class A(metaclass=M):
    pass


class A(B, metaclass=M):
    pass


class A(*t):
    pass


class A(B, *t):
    pass


class A(**kw):
    pass


class A(B, **kw):
    pass



""""Conftest for pure python parser."""
from pathlib import Path

import pytest
from pegen.build import build_parser
from pegen.utils import generate_parser


@pytest.fixture(scope="session")
def python_parser_cls():
    grammar_path = Path(__file__).parent.parent.parent / "data/python.gram"
    grammar = build_parser(grammar_path)[0]
    source_path = str(Path(__file__).parent / "parser_cache" / "py_parser.py")
    parser_cls = generate_parser(grammar, source_path, "PythonParser")

    return parser_cls


@pytest.fixture(scope="session")
def python_parse_file():
    grammar_path = Path(__file__).parent.parent.parent / "data/python.gram"
    grammar = build_parser(grammar_path)[0]
    source_path = str(Path(__file__).parent / "parser_cache" / "py_parser.py")
    parser_cls = generate_parser(grammar, source_path, "parse_file")

    return parser_cls


@pytest.fixture(scope="session")
def python_parse_str():
    grammar_path = Path(__file__).parent.parent.parent / "data/python.gram"
    grammar = build_parser(grammar_path)[0]
    source_path = str(Path(__file__).parent / "parser_cache" / "py_parser.py")
    parser_cls = generate_parser(grammar, source_path, "parse_string")

    return parser_cls



import argparse
import ast
import sys
import time
import token
import tokenize
import traceback
from abc import abstractmethod
from typing import Any, Callable, ClassVar, Dict, Optional, Tuple, Type, TypeVar, cast

from pegen.tokenizer import Mark, Tokenizer, exact_token_types

T = TypeVar("T")
P = TypeVar("P", bound="Parser")
F = TypeVar("F", bound=Callable[..., Any])

# Tokens added in Python 3.12
FSTRING_START = getattr(token, "FSTRING_START", None)
FSTRING_MIDDLE = getattr(token, "FSTRING_MIDDLE", None)
FSTRING_END = getattr(token, "FSTRING_END", None)


def logger(method: F) -> F:
    """For non-memoized functions that we want to be logged.

    (In practice this is only non-leader left-recursive functions.)
    """
    method_name = method.__name__

    def logger_wrapper(self: P, *args: object) -> F:
        if not self._verbose:
            return method(self, *args)
        argsr = ",".join(repr(arg) for arg in args)
        fill = "  " * self._level
        print(f"{fill}{method_name}({argsr}) .... (looking at {self.showpeek()})")
        self._level += 1
        tree = method(self, *args)
        self._level -= 1
        print(f"{fill}... {method_name}({argsr}) --> {tree!s:.200}")
        return tree

    logger_wrapper.__wrapped__ = method  # type: ignore
    return cast(F, logger_wrapper)


def memoize(method: F) -> F:
    """Memoize a symbol method."""
    method_name = method.__name__

    def memoize_wrapper(self: P, *args: object) -> F:
        mark = self._mark()
        key = mark, method_name, args
        # Fast path: cache hit, and not verbose.
        if key in self._cache and not self._verbose:
            tree, endmark = self._cache[key]
            self._reset(endmark)
            return tree
        # Slow path: no cache hit, or verbose.
        verbose = self._verbose
        argsr = ",".join(repr(arg) for arg in args)
        fill = "  " * self._level
        if key not in self._cache:
            if verbose:
                print(f"{fill}{method_name}({argsr}) ... (looking at {self.showpeek()})")
            self._level += 1
            tree = method(self, *args)
            self._level -= 1
            if verbose:
                print(f"{fill}... {method_name}({argsr}) -> {tree!s:.200}")
            endmark = self._mark()
            self._cache[key] = tree, endmark
        else:
            tree, endmark = self._cache[key]
            if verbose:
                print(f"{fill}{method_name}({argsr}) -> {tree!s:.200}")
            self._reset(endmark)
        return tree

    memoize_wrapper.__wrapped__ = method  # type: ignore
    return cast(F, memoize_wrapper)


def memoize_left_rec(method: Callable[[P], Optional[T]]) -> Callable[[P], Optional[T]]:
    """Memoize a left-recursive symbol method."""
    method_name = method.__name__

    def memoize_left_rec_wrapper(self: P) -> Optional[T]:
        mark = self._mark()
        key = mark, method_name, ()
        # Fast path: cache hit, and not verbose.
        if key in self._cache and not self._verbose:
            tree, endmark = self._cache[key]
            self._reset(endmark)
            return tree
        # Slow path: no cache hit, or verbose.
        verbose = self._verbose
        fill = "  " * self._level
        if key not in self._cache:
            if verbose:
                print(f"{fill}{method_name} ... (looking at {self.showpeek()})")
            self._level += 1

            # For left-recursive rules we manipulate the cache and
            # loop until the rule shows no progress, then pick the
            # previous result.  For an explanation why this works, see
            # https://github.com/PhilippeSigaud/Pegged/wiki/Left-Recursion
            # (But we use the memoization cache instead of a static
            # variable; perhaps this is similar to a paper by Warth et al.
            # (http://web.cs.ucla.edu/~todd/research/pub.php?id=pepm08).

            # Prime the cache with a failure.
            self._cache[key] = None, mark
            lastresult, lastmark = None, mark
            depth = 0
            if verbose:
                print(f"{fill}Recursive {method_name} at {mark} depth {depth}")

            while True:
                self._reset(mark)
                self.in_recursive_rule += 1
                try:
                    result = method(self)
                finally:
                    self.in_recursive_rule -= 1
                endmark = self._mark()
                depth += 1
                if verbose:
                    print(
                        f"{fill}Recursive {method_name} at {mark} depth {depth}: {result!s:.200} to {endmark}"
                    )
                if not result:
                    if verbose:
                        print(f"{fill}Fail with {lastresult!s:.200} to {lastmark}")
                    break
                if endmark <= lastmark:
                    if verbose:
                        print(f"{fill}Bailing with {lastresult!s:.200} to {lastmark}")
                    break
                self._cache[key] = lastresult, lastmark = result, endmark

            self._reset(lastmark)
            tree = lastresult

            self._level -= 1
            if verbose:
                print(f"{fill}{method_name}() -> {tree!s:.200} [cached]")
            if tree:
                endmark = self._mark()
            else:
                endmark = mark
                self._reset(endmark)
            self._cache[key] = tree, endmark
        else:
            tree, endmark = self._cache[key]
            if verbose:
                print(f"{fill}{method_name}() -> {tree!s:.200} [fresh]")
            if tree:
                self._reset(endmark)
        return tree

    memoize_left_rec_wrapper.__wrapped__ = method  # type: ignore
    return memoize_left_rec_wrapper


class Parser:
    """Parsing base class."""

    KEYWORDS: ClassVar[Tuple[str, ...]]

    SOFT_KEYWORDS: ClassVar[Tuple[str, ...]]

    def __init__(self, tokenizer: Tokenizer, *, verbose: bool = False):
        self._tokenizer = tokenizer
        self._verbose = verbose
        self._level = 0
        self._cache: Dict[Tuple[Mark, str, Tuple[Any, ...]], Tuple[Any, Mark]] = {}

        # Integer tracking wether we are in a left recursive rule or not. Can be useful
        # for error reporting.
        self.in_recursive_rule = 0

        # Pass through common tokenizer methods.
        self._mark = self._tokenizer.mark
        self._reset = self._tokenizer.reset

        # Are we looking for syntax error ? When true enable matching on invalid rules
        self.call_invalid_rules = False

    @abstractmethod
    def start(self) -> Any:
        """Expected grammar entry point.

        This is not strictly necessary but is assumed to exist in most utility
        functions consuming parser instances.

        """
        pass

    def showpeek(self) -> str:
        tok = self._tokenizer.peek()
        return f"{tok.start[0]}.{tok.start[1]}: {token.tok_name[tok.type]}:{tok.string!r}"

    @memoize
    def name(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == token.NAME and tok.string not in self.KEYWORDS:
            return self._tokenizer.getnext()
        return None

    @memoize
    def number(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == token.NUMBER:
            return self._tokenizer.getnext()
        return None

    @memoize
    def string(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == token.STRING:
            return self._tokenizer.getnext()
        return None

    @memoize
    def fstring_start(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == FSTRING_START:
            return self._tokenizer.getnext()
        return None

    @memoize
    def fstring_middle(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == FSTRING_MIDDLE:
            return self._tokenizer.getnext()
        return None

    @memoize
    def fstring_end(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == FSTRING_END:
            return self._tokenizer.getnext()
        return None

    @memoize
    def op(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == token.OP:
            return self._tokenizer.getnext()
        return None

    @memoize
    def type_comment(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == token.TYPE_COMMENT:
            return self._tokenizer.getnext()
        return None

    @memoize
    def soft_keyword(self) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.type == token.NAME and tok.string in self.SOFT_KEYWORDS:
            return self._tokenizer.getnext()
        return None

    @memoize
    def expect(self, type: str) -> Optional[tokenize.TokenInfo]:
        tok = self._tokenizer.peek()
        if tok.string == type:
            return self._tokenizer.getnext()
        if type in exact_token_types:
            if tok.type == exact_token_types[type]:
                return self._tokenizer.getnext()
        if type in token.__dict__:
            if tok.type == token.__dict__[type]:
                return self._tokenizer.getnext()
        if tok.type == token.OP and tok.string == type:
            return self._tokenizer.getnext()
        return None

    def expect_forced(self, res: Any, expectation: str) -> Optional[tokenize.TokenInfo]:
        if res is None:
            raise self.make_syntax_error(f"expected {expectation}")
        return res

    def positive_lookahead(self, func: Callable[..., T], *args: object) -> T:
        mark = self._mark()
        ok = func(*args)
        self._reset(mark)
        return ok

    def negative_lookahead(self, func: Callable[..., object], *args: object) -> bool:
        mark = self._mark()
        ok = func(*args)
        self._reset(mark)
        return not ok

    def make_syntax_error(self, message: str, filename: str = "<unknown>") -> SyntaxError:
        tok = self._tokenizer.diagnose()
        return SyntaxError(message, (filename, tok.start[0], 1 + tok.start[1], tok.line))


def simple_parser_main(parser_class: Type[Parser]) -> None:
    argparser = argparse.ArgumentParser()
    argparser.add_argument(
        "-v",
        "--verbose",
        action="count",
        default=0,
        help="Print timing stats; repeat for more debug output",
    )
    argparser.add_argument(
        "-q", "--quiet", action="store_true", help="Don't print the parsed program"
    )
    argparser.add_argument("-r", "--run", action="store_true", help="Run the parsed program")
    argparser.add_argument("filename", help="Input file ('-' to use stdin)")

    args = argparser.parse_args()
    verbose = args.verbose
    verbose_tokenizer = verbose >= 3
    verbose_parser = verbose == 2 or verbose >= 4

    t0 = time.time()

    filename = args.filename
    if filename == "" or filename == "-":
        filename = "<stdin>"
        file = sys.stdin
    else:
        file = open(args.filename)
    try:
        tokengen = tokenize.generate_tokens(file.readline)
        tokenizer = Tokenizer(tokengen, verbose=verbose_tokenizer)
        parser = parser_class(tokenizer, verbose=verbose_parser)
        tree = parser.start()
        try:
            if file.isatty():
                endpos = 0
            else:
                endpos = file.tell()
        except IOError:
            endpos = 0
    finally:
        if file is not sys.stdin:
            file.close()

    t1 = time.time()

    if not tree:
        err = parser.make_syntax_error(filename)
        traceback.print_exception(err.__class__, err, None)
        sys.exit(1)

    if not args.quiet:
        print(ast.dump(tree))
    if args.run:
        exec(compile(tree, filename=filename, mode="exec"))

    if verbose:
        dt = t1 - t0
        diag = tokenizer.diagnose()
        nlines = diag.end[0]
        if diag.type == token.ENDMARKER:
            nlines -= 1
        print(f"Total time: {dt:.3f} sec; {nlines} lines", end="")
        if endpos:
            print(f" ({endpos} bytes)", end="")
        if dt:
            print(f"; {nlines / dt:.0f} lines/sec")
        else:
            print()
        print("Caches sizes:")
        print(f"  token array : {len(tokenizer._tokens):10}")
        print(f"        cache : {len(parser._cache):10}")
        ## print_memstats()



print("import: 'pegen'")
import pegen

print("import: 'pegen.grammar'")
import pegen.grammar

print("import: 'pegen.grammar_parser'")
import pegen.grammar_parser

print("import: 'pegen.grammar_visualizer'")
import pegen.grammar_visualizer

print("import: 'pegen.parser'")
import pegen.parser

print("import: 'pegen.parser_generator'")
import pegen.parser_generator

print("import: 'pegen.python_generator'")
import pegen.python_generator

print("import: 'pegen.tokenizer'")
import pegen.tokenizer

print("import: 'pegen.utils'")
import pegen.utils




import test
import a, b
import test as t
import test as t, y
import test.a
import test.b as b


from test import a
from test import a, b
from test import (
    a,
    b,
)
from test import a as b
from test import a as b, c
from test import a as b, c as d
from test import *
from test.a import b
from test.a import b as c
from test.a import b, c
from test.a import b as c, d


from . import a
from ... import b
from .... import c
from ..a import b
from ...a import c
from ....a import c
from . import a, b
from ..a import b, c
from ...a import c, d
from ....a import c, d



from typing import Any

from pegen.grammar import GrammarVisitor
from pegen.grammar_parser import GeneratedParser as GrammarParser
from pegen.utils import parse_string


class Visitor(GrammarVisitor):
    def __init__(self) -> None:
        self.n_nodes = 0

    def visit(self, node: Any, *args: Any, **kwargs: Any) -> None:
        self.n_nodes += 1
        super().visit(node, *args, **kwargs)


def test_parse_trivial_grammar() -> None:
    grammar = """
    start: 'a'
    """
    rules = parse_string(grammar, GrammarParser)
    visitor = Visitor()

    visitor.visit(rules)

    assert visitor.n_nodes == 6


def test_parse_or_grammar() -> None:
    grammar = """
    start: rule
    rule: 'a' | 'b'
    """
    rules = parse_string(grammar, GrammarParser)
    visitor = Visitor()

    visitor.visit(rules)

    # Grammar/Rule/Rhs/Alt/NamedItem/NameLeaf   -> 6
    #         Rule/Rhs/                         -> 2
    #                  Alt/NamedItem/StringLeaf -> 3
    #                  Alt/NamedItem/StringLeaf -> 3

    assert visitor.n_nodes == 14


def test_parse_repeat1_grammar() -> None:
    grammar = """
    start: 'a'+
    """
    rules = parse_string(grammar, GrammarParser)
    visitor = Visitor()

    visitor.visit(rules)

    # Grammar/Rule/Rhs/Alt/NamedItem/Repeat1/StringLeaf -> 6
    assert visitor.n_nodes == 7


def test_parse_repeat0_grammar() -> None:
    grammar = """
    start: 'a'*
    """
    rules = parse_string(grammar, GrammarParser)
    visitor = Visitor()

    visitor.visit(rules)

    # Grammar/Rule/Rhs/Alt/NamedItem/Repeat0/StringLeaf -> 6

    assert visitor.n_nodes == 7


def test_parse_optional_grammar() -> None:
    grammar = """
    start: 'a' ['b']
    """
    rules = parse_string(grammar, GrammarParser)
    visitor = Visitor()

    visitor.visit(rules)

    # Grammar/Rule/Rhs/Alt/NamedItem/StringLeaf                       -> 6
    #                      NamedItem/Opt/Rhs/Alt/NamedItem/Stringleaf -> 6

    assert visitor.n_nodes == 12



import textwrap
from typing import List

from pegen.grammar_parser import GeneratedParser as GrammarParser
from pegen.grammar_visualizer import ASTGrammarPrinter
from pegen.utils import parse_string


def test_simple_rule() -> None:
    grammar = """
    start: 'a' 'b'
    """
    rules = parse_string(grammar, GrammarParser)

    printer = ASTGrammarPrinter()
    lines: List[str] = []
    printer.print_grammar_ast(rules, printer=lines.append)

    output = "\n".join(lines)
    expected_output = textwrap.dedent(
        """\
    └──Rule
       └──Rhs
          └──Alt
             ├──NamedItem
             │  └──StringLeaf("'a'")
             └──NamedItem
                └──StringLeaf("'b'")
    """
    )

    assert output == expected_output


def test_multiple_rules() -> None:
    grammar = """
    start: a b
    a: 'a'
    b: 'b'
    """
    rules = parse_string(grammar, GrammarParser)

    printer = ASTGrammarPrinter()
    lines: List[str] = []
    printer.print_grammar_ast(rules, printer=lines.append)

    output = "\n".join(lines)
    expected_output = textwrap.dedent(
        """\
    └──Rule
       └──Rhs
          └──Alt
             ├──NamedItem
             │  └──NameLeaf('a')
             └──NamedItem
                └──NameLeaf('b')

    └──Rule
       └──Rhs
          └──Alt
             └──NamedItem
                └──StringLeaf("'a'")

    └──Rule
       └──Rhs
          └──Alt
             └──NamedItem
                └──StringLeaf("'b'")
                    """
    )

    assert output == expected_output


def test_deep_nested_rule() -> None:
    grammar = """
    start: 'a' ['b'['c'['d']]]
    """
    rules = parse_string(grammar, GrammarParser)

    printer = ASTGrammarPrinter()
    lines: List[str] = []
    printer.print_grammar_ast(rules, printer=lines.append)

    output = "\n".join(lines)
    print()
    print(output)
    expected_output = textwrap.dedent(
        """\
    └──Rule
       └──Rhs
          └──Alt
             ├──NamedItem
             │  └──StringLeaf("'a'")
             └──NamedItem
                └──Opt
                   └──Rhs
                      └──Alt
                         ├──NamedItem
                         │  └──StringLeaf("'b'")
                         └──NamedItem
                            └──Opt
                               └──Rhs
                                  └──Alt
                                     ├──NamedItem
                                     │  └──StringLeaf("'c'")
                                     └──NamedItem
                                        └──Opt
                                           └──Rhs
                                              └──Alt
                                                 └──NamedItem
                                                    └──StringLeaf("'d'")
                            """
    )

    assert output == expected_output



pass
pass;

assert a
assert a; assert b
assert a, "eee"

raise RuntimeError
raise RuntimeError from e

return
return 1
return 1,
return *a

del a
del (a)
del a, b,
del a[:]
del a.b
del (a,)
del (a, b)
del [a, b]
del a;

global a
global a, b
nonlocal a
nonlocal a, b

yield a
yield from a


for i in a:
    pass

for i, in a:
    pass

for (i,) in a:
    pass

for (i,), in a:
    pass

for i, *j in a:
    pass

for i, (a, *b) in a:
    pass

async for i in a:
    pass

async for i, in a:
    pass

async for (i,) in a:
    pass

async for (i,), in a:
    pass

async for i, *j in a:
    pass

async for i, (a, *b) in a:
    pass

for i in b:
    pass
else:
    pass


if a:
    b=1

if a:
    pass
else:
    pass

if a:
    pass
elif b:
    pass
else:
    pass

if a:
    pass
elif b:
    pass
elif c:
    pass


while s:
    pass

while False:
    pass
else:
    pass


for i in a:
    continue

for i in a:
    break


with a:
    pass

with a, b:
    pass

with a as b:
    pass

with a as b, c:
    pass

async with a:
    pass

async with a, b:
    pass

async with a as b:
    pass

async with a as b, c:
    pass


try:
    pass
finally:
    pass


try:
    pass
except:
    raise
finally:
    pass

try:
    pass
except ValueError:
    pass
except (IndexError, RuntimeError,):
    pass
except Exception as e:
    pass
else:
    pass
finally:
    pass



import io
import sys
from tokenize import NEWLINE, NUMBER, ENDMARKER, TokenInfo, generate_tokens

from pegen.tokenizer import Tokenizer


def test_peek_getnext():
    source = io.StringIO("# test\n1")
    t = Tokenizer(generate_tokens(source.readline))
    assert t.peek() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1")
    assert t.getnext() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1")
    assert t.peek() == TokenInfo(
        NEWLINE, "", (2, 1), (2, 2), "1" if sys.version_info >= (3, 12) else ""
    )
    assert t.getnext() == TokenInfo(
        NEWLINE, "", (2, 1), (2, 2), "1" if sys.version_info >= (3, 12) else ""
    )


def test_mark_reset():
    source = io.StringIO("\n1 2")
    t = Tokenizer(generate_tokens(source.readline))
    index = t.mark()
    assert t.peek() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1 2")
    assert t.getnext() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1 2")
    t.reset(index)
    assert t.peek() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1 2")
    assert t.getnext() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1 2")


def test_last_non_whitespace():
    source = io.StringIO("\n1\n2")
    t = Tokenizer(generate_tokens(source.readline))
    assert t.peek() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1\n")
    assert t.getnext() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1\n")
    assert t.getnext() == TokenInfo(NEWLINE, "\n", (2, 1), (2, 2), "1\n")
    assert t.get_last_non_whitespace_token() == TokenInfo(NUMBER, "1", (2, 0), (2, 1), "1\n")


def test_get_lines():
    source = io.StringIO("1\n2\n3")
    t = Tokenizer(generate_tokens(source.readline))
    while True:
        if t.getnext().type == ENDMARKER:
            break
    assert t.get_lines([1, 2, 3]) == ["1\n", "2\n", "3"]



match 0:
    case 0:
        x = True


match 0:
    case 0 if False:
        x = False
    case 0 if True:
        x = True


match 0:
    case 0:
        x = True
    case 0:
        x = False


x = False
match 0:
    case 0 | 1 | 2 | 3:
        x = True


x = False
match 1:
    case 0 | 1 | 2 | 3:
        x = True


x = False
match 2:
    case 0 | 1 | 2 | 3:
        x = True


x = False
match 3:
    case 0 | 1 | 2 | 3:
        x = True


x = False
match 4:
    case 0 | 1 | 2 | 3:
        x = True


x = 0
class A:
    y = 1
match x:
    case A.y as z:
        pass


class A:
    B = 0
match 0:
    case x if x:
        z = 0
    case _ as y if y == x and y:
        z = 1
    case A.B:
        z = 2


match ():
    case []:
        x = 0


match (0, 1, 2):
    case [*x]:
        y = 0


match (0, 1, 2):
    case [0, *x]:
        y = 0


match (0, 1, 2):
    case [0, 1, *x,]:
        y = 0


match (0, 1, 2):
    case [0, 1, 2, *x]:
        y = 0


match (0, 1, 2):
    case [*x, 2,]:
        y = 0


match (0, 1, 2):
    case [*x, 1, 2]:
        y = 0


match (0, 1, 2):
    case [*x, 0, 1, 2,]:
        y = 0


match (0, 1, 2):
    case [0, *x, 2]:
        y = 0


match (0, 1, 2):
    case [0, 1, *x, 2,]:
        y = 0


match (0, 1, 2):
    case [0, *x, 1, 2]:
        y = 0


match (0, 1, 2):
    case [*x,]:
        y = 0


x = {}
match x:
    case {}:
        y = 0


x = {0: 0}
match x:
    case {}:
        y = 0


x = {}
y = None
match x:
    case {0: 0}:
        y = 0


x = {0: 0}
match x:
    case {0: (0 | 1 | 2 as z)}:
        y = 0


x = {0: 1}
match x:
    case {0: (0 | 1 | 2 as z)}:
        y = 0


x = {0: 2}
match x:
    case {0: (0 | 1 | 2 as z)}:
        y = 0


x = {0: 3}
y = None
match x:
    case {0: (0 | 1 | 2 as z)}:
        y = 0


x = {}
y = None
match x:
    case {0: [1, 2, {}]}:
        y = 0
    case {0: [1, 2, {}], 1: [[]]}:
        y = 1
    case []:
        y = 2


x = {False: (True, 2.0, {})}
match x:
    case {0: [1, 2, {}]}:
        y = 0
    case {0: [1, 2, {}], 1: [[]]}:
        y = 1
    case []:
        y = 2


x = {False: (True, 2.0, {}), 1: [[]], 2: 0}
match x:
    case {0: [1, 2, {}]}:
        y = 0
    case {0: [1, 2, {}], 1: [[]]}:
        y = 1
    case []:
        y = 2


x = {False: (True, 2.0, {}), 1: [[]], 2: 0}
match x:
    case {0: [1, 2]}:
        y = 0
    case {0: [1, 2, {}], 1: [[]]}:
        y = 1
    case []:
        y = 2


x = []
match x:
    case {0: [1, 2, {}]}:
        y = 0
    case {0: [1, 2, {}], 1: [[]]}:
        y = 1
    case []:
        y = 2


x = {0: 0}
match x:
    case {0: [1, 2, {}]}:
        y = 0
    case {0: ([1, 2, {}] | False)} | {1: [[]]} | {0: [1, 2, {}]} | [] | "X" | {}:
        y = 1
    case []:
        y = 2


x = {0: 0}
match x:
    case {0: [1, 2, {}]}:
        y = 0
    case {0: [1, 2, {}] | True} | {1: [[]]} | {0: [1, 2, {}]} | [] | "X" | {}:
        y = 1
    case []:
        y = 2

x = {0: 0}
match x:
    case {None: 1}:
        y = 0
    case {True: 1}:
        y = 0
    case {False: 1}:
        y = 0
    case {1 + 1j: 1}:
        y = 0
    case {a.b: 1}:
        y = 0


x = 0
match x:
    case 0 | 1 | 2:
        y = 0


x = 1
match x:
    case 0 | 1 | 2:
        y = 0


x = 2
match x:
    case 0 | 1 | 2:
        y = 0


x = 3
y = None
match x:
    case 0 | 1 | 2:
        y = 0


x = 0
match x:
    case (0 as z) | (1 as z) | (2 as z) if z == x % 2:
        y = 0


x = 1
match x:
    case (0 as z) | (1 as z) | (2 as z) if z == x % 2:
        y = 0


x = 2
y = None
match x:
    case (0 as z) | (1 as z) | (2 as z) if z == x % 2:
        y = 0


x = 3
y = None
match x:
    case (0 as z) | (1 as z) | (2 as z) if z == x % 2:
        y = 0


x = ()
match x:
    case []:
        y = 0


x = ()
match x:
    case ():
        y = 0


x = (0,)
match x:
    case [0]:
        y = 0


x = ((),)
match x:
    case [[]]:
        y = 0


x = [0, 1]
match x:
    case [0, 1] | [1, 0]:
        y = 0


x = [1, 0]
match x:
    case [0, 1] | [1, 0]:
        y = 0


x = [0, 0]
y = None
match x:
    case [0, 1] | [1, 0]:
        y = 0


w = None
x = [1, 0]
match x:
    case [(0 as w)]:
        y = 0
    case [z] | [1, (0 | 1 as z)] | [z]:
        y = 1


x = [1, 0]
match x:
    case [0]:
        y = 0
    case [1, 0] if (x := x[:0]):
        y = 1
    case [1, 0]:
        y = 2


x = {0}
y = None
match x:
    case [0]:
        y = 0


x = set()
y = None
match x:
    case []:
        y = 0


x = iter([1, 2, 3])
y = None
match x:
    case []:
        y = 0


x = {}
y = None
match x:
    case []:
        y = 0


x = {0: False, 1: True}
y = None
match x:
    case [0, 1]:
        y = 0


x = 0
match x:
    case 0:
        y = 0


x = 0
y = None
match x:
    case False:
        y = 0


x = 0
y = None
match x:
    case 1:
        y = 0


x = 0
y = None
match x:
    case None:
        y = 0


x = 0
match x:
    case 0:
        y = 0
    case 0:
        y = 1


x = 0
y = None
match x:
    case 1:
        y = 0
    case 1:
        y = 1


x = "x"
match x:
    case "x":
        y = 0
    case "y":
        y = 1


x = "x"
match x:
    case "y":
        y = 0
    case "x":
        y = 1


x = "x"
match x:
    case "":
        y = 0
    case "x":
        y = 1


x = b"x"
match x:
    case b"y":
        y = 0
    case b"x":
        y = 1


x = 0
match x:
    case 0 if False:
        y = 0
    case 0:
        y = 1


x = 0
y = None
match x:
    case 0 if 0:
        y = 0
    case 0 if 0:
        y = 1


x = 0
match x:
    case 0 if True:
        y = 0
    case 0 if True:
        y = 1


x = 0
match x:
    case 0 if 1:
        y = 0
    case 0 if 1:
        y = 1


x = 0
match x:
    case 0 if True:
        y = 0
    case 0 if True:
        y = 1
y = 2


x = 0
match x:
    case 0 if 0:
        y = 0
    case 0 if 1:
        y = 1
y = 2


x = 0
y = None
match x:
    case 0 if not (x := 1):
        y = 0
    case 1:
        y = 1


x = "x"
match x:
    case ["x"]:
        y = 0
    case "x":
        y = 1


x = b"x"
match x:
    case [b"x"]:
        y = 0
    case ["x"]:
        y = 1
    case [120]:
        y = 2
    case b"x":
        y = 4


x = bytearray(b"x")
y = None
match x:
    case [120]:
        y = 0
    case 120:
        y = 1


x = ""
match x:
    case []:
        y = 0
    case [""]:
        y = 1
    case "":
        y = 2


x = "xxx"
match x:
    case ["x", "x", "x"]:
        y = 0
    case ["xxx"]:
        y = 1
    case "xxx":
        y = 2


x = b"xxx"
match x:
    case [120, 120, 120]:
        y = 0
    case [b"xxx"]:
        y = 1
    case b"xxx":
        y = 2


x = 0
match x:
    case 0 if not (x := 1):
        y = 0
    case (0 as z):
        y = 1


x = 0
match x:
    case (1 as z) if not (x := 1):
        y = 0
    case 0:
        y = 1


x = 0
match x:
    case (0 as z):
        y = 0


x = 0
y = None
match x:
    case (1 as z):
        y = 0


x = 0
y = None
match x:
    case (0 as z) if (w := 0):
        y = 0


x = 0
match x:
    case ((0 as w) as z):
        y = 0


x = 0
match x:
    case (0 | 1) | 2:
        y = 0


x = 1
match x:
    case (0 | 1) | 2:
        y = 0


x = 2
match x:
    case (0 | 1) | 2:
        y = 0


x = 3
y = None
match x:
    case (0 | 1) | 2:
        y = 0


x = 0
match x:
    case 0 | (1 | 2):
        y = 0


x = 1
match x:
    case 0 | (1 | 2):
        y = 0


x = 2
match x:
    case 0 | (1 | 2):
        y = 0


x = 3
y = None
match x:
    case 0 | (1 | 2):
        y = 0


x = 0
match x:
    case -0:
        y = 0


x = 0
match x:
    case -0.0:
        y = 0


x = 0
match x:
    case -0j:
        y = 0


x = 0
match x:
    case -0.0j:
        y = 0


x = -1
match x:
    case -1:
        y = 0


x = -1.5
match x:
    case -1.5:
        y = 0


x = -1j
match x:
    case -1j:
        y = 0


x = -1.5j
match x:
    case -1.5j:
        y = 0


x = 0
match x:
    case 0 + 0j:
        y = 0


x = 0
match x:
    case 0 - 0j:
        y = 0


x = 0
match x:
    case -0 + 0j:
        y = 0


x = 0
match x:
    case -0 - 0j:
        y = 0


x = 0.25 + 1.75j
match x:
    case 0.25 + 1.75j:
        y = 0


x = 0.25 - 1.75j
match x:
    case 0.25 - 1.75j:
        y = 0


x = -0.25 + 1.75j
match x:
    case -0.25 + 1.75j:
        y = 0


x = -0.25 - 1.75j
match x:
    case -0.25 - 1.75j:
        y = 0


class A:
    B = 0
x = 0
match x:
    case A.B:
        y = 0


class A:
    class B:
        C = 0
x = 0
match x:
    case A.B.C:
        y = 0


class A:
    class B:
        C = 0
        D = 1
x = 1
match x:
    case A.B.C:
        y = 0
    case A.B.D:
        y = 1


class A:
    class B:
        class C:
            D = 0
x = 0
match x:
    case A.B.C.D:
        y = 0


class A:
    class B:
        class C:
            D = 0
            E = 1
x = 1
match x:
    case A.B.C.D:
        y = 0
    case A.B.C.E:
        y = 1


match = case = 0
match match:
    case case:
        x = 0


match = case = 0
match case:
    case match:
        x = 0


x = []
match x:
    case [*_, _]:
        y = 0
    case []:
        y = 1


x = collections.defaultdict(int)
match x:
    case {0: 0}:
        y = 0
    case {}:
        y = 1


x = collections.defaultdict(int)
match x:
    case {0: 0}:
        y = 0
    case {**z}:
        y = 1


match ():
    case ():
        x = 0


match (0, 1, 2):
    case (*x,):
        y = 0


match (0, 1, 2):
    case 0, *x:
        y = 0


match (0, 1, 2):
    case (0, 1, *x,):
        y = 0


match (0, 1, 2):
    case 0, 1, 2, *x:
        y = 0


match (0, 1, 2):
    case *x, 2,:
        y = 0


match (0, 1, 2):
    case (*x, 1, 2):
        y = 0


match (0, 1, 2):
    case *x, 0, 1, 2,:
        y = 0


match (0, 1, 2):
    case (0, *x, 2):
        y = 0


match (0, 1, 2):
    case 0, 1, *x, 2,:
        y = 0


match (0, 1, 2):
    case (0, *x, 1, 2):
        y = 0


match (0, 1, 2):
    case *x,:
        y = 0


x = collections.defaultdict(int, {0: 1})
match x:
    case {1: 0}:
        y = 0
    case {0: 0}:
        y = 1
    case {}:
        y = 2


x = collections.defaultdict(int, {0: 1})
match x:
    case {1: 0}:
        y = 0
    case {0: 0}:
        y = 1
    case {**z}:
        y = 2


x = collections.defaultdict(int, {0: 1})
match x:
    case {1: 0}:
        y = 0
    case {0: 0}:
        y = 1
    case {0: _, **z}:
        y = 2


x = {0: 1}
match x:
    case {1: 0}:
        y = 0
    case {0: 0}:
        y = 0
    case {}:
        y = 1


x = {0: 1}
match x:
    case {1: 0}:
        y = 0
    case {0: 0}:
        y = 0
    case {**z}:
        y = 1


x = {0: 1}
match x:
    case {1: 0}:
        y = 0
    case {0: 0}:
        y = 0
    case {0: _, **z}:
        y = 1


x = False
match x:
    case bool(z):
        y = 0


x = True
match x:
    case bool(z):
        y = 0


x = bytearray()
match x:
    case bytearray(z):
        y = 0


x = b""
match x:
    case bytes(z):
        y = 0


x = {}
match x:
    case dict(z):
        y = 0


x = 0.0
match x:
    case float(z):
        y = 0


x = frozenset()
match x:
    case frozenset(z):
        y = 0


x = 0
match x:
    case int(z):
        y = 0


x = []
match x:
    case list(z):
        y = 0


x = set()
match x:
    case set(z):
        y = 0


x = ""
match x:
    case str(z):
        y = 0


x = ()
match x:
    case tuple(z):
        y = 0


x = 0
match x,:
    case y,:
        z = 0


w = 0
x = 0
match w, x:
    case y, z:
        v = 0


x = 0
match w := x,:
    case y as v,:
        z = 0


x = 0
y = None
match x:
    case 0 if x:
        y = 0


x = 0
y = None
match x:
    case 1e1000:
        y = 0


x = 0
match x:
    case z:
        y = 0


x = 0
y = None
match x:
    case _ if x:
        y = 0


x = 0
match x:
    case -1e1000:
        y = 0
    case 0:
        y = 1


x = 0
match x:
    case 0 if not x:
        y = 0
    case 1:
        y = 1


x = 0
z = None
match x:
    case 0:
        y = 0
    case z if x:
        y = 1


x = 0
match x:
    case 0:
        y = 0
    case _:
        y = 1


x = 0
match x:
    case 1 if x:
        y = 0
    case 0:
        y = 1


x = 0
y = None
match x:
    case 1:
        y = 0
    case 1 if not x:
        y = 1


x = 0
match x:
    case 1:
        y = 0
    case z:
        y = 1


x = 0
match x:
    case 1 if x:
        y = 0
    case _:
        y = 1


x = 0
match x:
    case z if not z:
        y = 0
    case 0 if x:
        y = 1


x = 0
match x:
    case z if not z:
        y = 0
    case 1:
        y = 1


x = 0
match x:
    case z if not x:
        y = 0
    case z:
        y = 1


x = 0
match x:
    case z if not z:
        y = 0
    case _ if x:
        y = 1


x = 0
match x:
    case _ if not x:
        y = 0
    case 0:
        y = 1


x = 0
y = None
match x:
    case _ if x:
        y = 0
    case 1:
        y = 1


x = 0
z = None
match x:
    case _ if not x:
        y = 0
    case z if not x:
        y = 1


x = 0
match x:
    case _ if not x:
        y = 0
    case _:
        y = 1


match status:
    case 400:
        return "Bad request"
    case 401:
        return "Unauthorized"
    case 403:
        return "Forbidden"
    case 404:
        return "Not found"
    case 418:
        return "I'm a teapot"
    case _:
        return "Something else"


match status:
    case 400:
        return "Bad request"
    case 401 | 403 | 404:
        return "Not allowed"
    case 418:
        return "I'm a teapot"


match point:
    case (0, 0):
        return "Origin"
    case (0, y):
        return f"Y={y}"
    case (x, 0):
        return f"X={x}"
    case (x, y):
        return f"X={x}, Y={y}"
    case _:
        raise ValueError("Not a point")


match point:
    case Point(0, 0):
        return "Origin"
    case Point(0, y):
        return f"Y={y}"
    case Point(x, 0):
        return f"X={x}"
    case Point():
        return "Somewhere else"
    case _:
        return "Not a point"


match point:
    case Point(1, var):
        return var


match point:
    case Point(1, y=var):
        return var


match point:
    case Point(x=1, y=var):
        return var


match point:
    case Point(y=var, x=1):
        return var


match points:
    case []:
        return "No points"
    case [Point(0, 0)]:
        return "The origin"
    case [Point(x, y)]:
        return f"Single point {x}, {y}"
    case [Point(0, y1), Point(0, y2)]:
        return f"Two on the Y axis at {y1}, {y2}"
    case _:
        return "Something else"


match point:
    case Point(x, y) if x == y:
        return f"Y=X at {x}"
    case Point(x, y):
        return "Not on the diagonal"


class Seq(collections.abc.Sequence):
    __getitem__ = None
    def __len__(self):
        return 0
match Seq():
    case []:
        y = 0


class Seq(collections.abc.Sequence):
    __getitem__ = None
    def __len__(self):
        return 42
match Seq():
    case [*_]:
        y = 0


class Seq(collections.abc.Sequence):
    def __getitem__(self, i):
        return i
    def __len__(self):
        return 42
match Seq():
    case [x, *_, y]:
        z = 0


w = range(10)
match w:
    case [x, y, *rest]:
        z = 0

w = range(100)
match w:
    case (x, y, *rest):
        z = 0


w = range(1000)
match w:
    case x, y, *rest:
        z = 0


w = range(1 << 10)
match w:
    case [x, y, *_]:
        z = 0


w = range(1 << 20)
match w:
    case (x, y, *_):
        z = 0


w = range(1 << 30)
match w:
    case x, y, *_:
        z = 0


x = {"bandwidth": 0, "latency": 1}
match x:
    case {"bandwidth": b, "latency": l}:
        y = 0


x = {"bandwidth": 0, "latency": 1, "key": "value"}
match x:
    case {"latency": l, "bandwidth": b}:
        y = 0


x = {"bandwidth": 0, "latency": 1, "key": "value"}
match x:
    case {"bandwidth": b, "latency": l, **rest}:
        y = 0


x = {"bandwidth": 0, "latency": 1}
match x:
    case {"latency": l, "bandwidth": b, **rest}:
        y = 0


w = [Point(-1, 0), Point(1, 2)]
match w:
    case (Point(x1, y1), Point(x2, y2) as p2):
        z = 0


class Color(enum.Enum):
    RED = 0
    GREEN = 1
    BLUE = 2

    match color:
        case Color.RED:
            return "I see red!"
        case Color.GREEN:
            return "Grass is green"
        case Color.BLUE:
            return "I'm feeling the blues :("


class Color(int, enum.Enum):
    RED = 0
    GREEN = 1
    BLUE = 2

    match color:
        case Color.RED:
            return "I see red!"
        case Color.GREEN:
            return "Grass is green"
        case Color.BLUE:
            return "I'm feeling the blues :("

class Class:
    __match_args__ = ("a", "b")
c = Class()
c.a = 0
c.b = 1
match c:
    case Class(x, y):
        z = 0

class Class:
    __match_args__ = ("a", "b")
c = Class()
c.a = 0
c.b = 1
match c:
    case Class(x, b=y):
        z = 0


class Parent:
    __match_args__ = "a", "b"
class Child(Parent):
    __match_args__ = ("c", "d")
c = Child()
c.a = 0
c.b = 1
match c:
    case Parent(x, y):
        z = 0


class Parent:
    __match_args__ = ("a", "b")
class Child(Parent):
    __match_args__ = "c", "d"
c = Child()
c.a = 0
c.b = 1
match c:
    case Parent(x, b=y):
        z = 0

match w:
    case 42:
        out = locals()
        del out["w"]
        return out


match w:
    case 42.0:
        out = locals()
        del out["w"]
        return out


match w:
    case 1 | 2 | 3:
        out = locals()
        del out["w"]
        return out

match w:
    case [1, 2] | [3, 4]:
        out = locals()
        del out["w"]
        return out

match w:
    case x:
        out = locals()
        del out["w"]
        return out

match w:
    case _:
        out = locals()
        del out["w"]
        return out

match w:
    case (x, y, z):
        out = locals()
        del out["w"]
        return out

match w:
    case {"x": x, "y": "y", "z": z}:
        out = locals()
        del out["w"]
        return out

match w:
    case MyClass(int(xx), y="hello"):
        out = locals()
        del out["w"]
        return out

match w:
    case (p, q) as x:
        out = locals()
        del out["w"]
        return out

match 42:
    case 42:
        return locals()

match 1:
    case 1 | 2 | 3:
        return locals()

match ...:
    case _:
        return locals()

match ...:
    case abc:
        return locals()

match ..., ...:
    case a, b:
        return locals()

match {"k": ..., "l": ...}:
    case {"k": a, "l": b}:
        return locals()

match MyClass(..., ...):
    case MyClass(x, y=y):
        return locals()

match ...:
    case b as a:
        return locals()

match x:
    case _:
        return 0

match x:
    case 0:
        return 0

match x:
    case 0:
        return 0
    case _:
        return 1

match x:
    case 0:
        return 0
    case 1:
        return 1

match x:
    case 0:
        return 0
    case 1:
        return 1
    case _:
        return 2

match x:
    case 0:
        return 0
    case 1:
        return 1
    case 2:
        return 2



async def f():
    pass


async def f():
    await b


async def f():
    async for i in range(10):
        pass


async def f():
    async with open(f) as p:
        pass


async def f():
    a = [i async for i in range(10)]
    return a
